import logging
import os
from pathlib import Path
from typing import Optional

from sciencebeam_parser.app.context import DownloadManagerProtocol
from sciencebeam_parser.lookup import MergedTextLookUp, SimpleTextLookUp, TextLookUp
from sciencebeam_parser.lookup.xml_lookup import load_xml_lookup_from_file


LOGGER = logging.getLogger(__name__)


def is_xml_filename(path: str):
    name = os.path.basename(path)
    return name.lower().endswith('.xml') or name.lower().endswith('.xml.gz')


def load_lookup_from_text_file(path: str) -> TextLookUp:
    return SimpleTextLookUp(
        set(Path(path).read_text(encoding='utf-8').splitlines()) - {''}
    )


def load_lookup_from_path(
    path: Optional[str],
    download_manager: DownloadManagerProtocol
) -> Optional[TextLookUp]:
    if not path:
        return None
    local_path = download_manager.download_if_url(path)
    LOGGER.info('loading lookup from: %r (%r)', path, local_path)
    if is_xml_filename(path):
        return load_xml_lookup_from_file(local_path)
    return load_lookup_from_text_file(local_path)


def load_lookup_from_config(
    config: Optional[dict],
    download_manager: DownloadManagerProtocol
) -> Optional[TextLookUp]:
    if not config:
        return None
    paths = config.get('paths', [])
    LOGGER.info('loading lookup from: %r', paths)
    return MergedTextLookUp([
        load_lookup_from_path(path, download_manager=download_manager)
        for path in paths
    ])



import logging

from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/TableParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = ['text', 'figure[@type="table"]']

# Note:
#   The table training data generation is different to figures in
#   how the following labels are mapped: `content`, `other`, `note`
TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<figure_head>': ROOT_TRAINING_XML_ELEMENT_PATH + ['head'],
    '<label>': ROOT_TRAINING_XML_ELEMENT_PATH + ['head', 'label'],
    '<figDesc>': ROOT_TRAINING_XML_ELEMENT_PATH + ['figDesc'],
    '<content>': ROOT_TRAINING_XML_ELEMENT_PATH + ['table'],
    '<other>': ROOT_TRAINING_XML_ELEMENT_PATH + ['other'],
    '<note>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note']
}


class TableTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.table.tei.xml'
    DEFAULT_DATA_FILENAME_SUFFIX = '.table'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            use_tei_namespace=False,
            root_tag='tei',
            default_tei_filename_suffix=(
                TableTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=(
                TableTeiTrainingDataGenerator.DEFAULT_DATA_FILENAME_SUFFIX
            ),
            default_tei_sub_directory='table/corpus/tei',
            default_data_sub_directory='table/corpus/raw'
        )


class TableTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=False
        )



import logging
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from pathlib import Path
from typing import Iterable, Iterator, List, Type, TypeVar

from flask import Blueprint, jsonify, request, Response, url_for
from flask.helpers import send_file
from lxml import etree

from sciencebeam_trainer_delft.sequence_labelling.reader import load_data_crf_lines

from sciencebeam_trainer_delft.sequence_labelling.tag_formatter import (
    TagOutputFormats,
    iter_format_tag_result
)
from werkzeug.exceptions import BadRequest

from sciencebeam_parser.app.parser import (
    BadRequestScienceBeamParserError,
    ScienceBeamParser,
    ScienceBeamParserSession,
    ScienceBeamParserSessionSource
)
from sciencebeam_parser.external.pdfalto.wrapper import PdfAltoWrapper
from sciencebeam_parser.external.pdfalto.parser import parse_alto_root
from sciencebeam_parser.models.data import AppFeaturesContext, DocumentFeaturesContext
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.document.layout_document import LayoutDocument
from sciencebeam_parser.document.semantic_document import (
    SemanticMixedContentWrapper,
    SemanticRawAffiliationAddress,
    SemanticRawAuthors,
    SemanticRawFigure,
    SemanticRawReference,
    SemanticRawReferenceText,
    SemanticRawTable,
    SemanticReference,
    T_SemanticContentWrapper,
    iter_by_semantic_type_recursively
)
from sciencebeam_parser.processors.fulltext.config import RequestFieldNames
from sciencebeam_parser.utils.data_wrapper import (
    get_data_wrapper_with_improved_media_type_or_filename
)
from sciencebeam_parser.utils.flask import (
    assert_and_get_first_accept_matching_media_type,
    get_bool_request_arg,
    get_int_request_arg,
    get_required_post_data,
    get_required_post_data_wrapper,
    get_str_request_arg
)
from sciencebeam_parser.utils.media_types import (
    MediaTypes
)
from sciencebeam_parser.utils.text import normalize_text, parse_comma_separated_value
from sciencebeam_parser.utils.tokenizer import get_tokenized_tokens
from sciencebeam_parser.processors.fulltext.api import (
    FullTextProcessorConfig
)


LOGGER = logging.getLogger(__name__)


T = TypeVar('T')


class RequestArgs:
    FIRST_PAGE = 'first_page'
    LAST_PAGE = 'last_page'
    OUTPUT_FORMAT = 'output_format'
    NO_USE_SEGMENTATION = 'no_use_segmentation'
    INCLUDES = 'includes'


class ModelOutputFormats:
    RAW_DATA = 'raw_data'


DEFAULT_MODEL_OUTPUT_FORMAT = TagOutputFormats.JSON
VALID_MODEL_OUTPUT_FORMATS = {
    ModelOutputFormats.RAW_DATA,
    TagOutputFormats.JSON,
    TagOutputFormats.DATA,
    TagOutputFormats.XML
}


def _get_file_upload_form(title: str):
    return (
        '''
        <!doctype html>
        <title>{title}</title>
        <h1>{title}</h1>
        <form target=_blank method=post enctype=multipart/form-data>
        <input type=file name=file>
        <input type=submit value=Upload>
        </form>
        '''
    ).format(title=title)


def normalize_and_tokenize_text(text: str) -> List[str]:
    return get_tokenized_tokens(
        normalize_text(text),
        keep_whitespace=True
    )


def normalize_layout_document(
    layout_document: LayoutDocument,
    **kwargs
) -> LayoutDocument:
    return (
        layout_document
        .retokenize(tokenize_fn=normalize_and_tokenize_text)
        .remove_empty_blocks(**kwargs)
    )


class ModelNestedBluePrint:
    def __init__(
        self,
        name: str,
        model: Model,
        pdfalto_wrapper: PdfAltoWrapper,
        app_features_context: AppFeaturesContext,
        model_name: str = 'dummy'
    ):
        self.name = name
        self.model = model
        self.pdfalto_wrapper = pdfalto_wrapper
        self.app_features_context = app_features_context
        self.model_name = model_name

    def add_routes(self, parent_blueprint: Blueprint, url_prefix: str):
        parent_blueprint.route(
            url_prefix, methods=['GET'], endpoint=f'{url_prefix}_get'
        )(self.handle_get)
        parent_blueprint.route(
            url_prefix, methods=['POST'], endpoint=f'{url_prefix}_post'
        )(self.handle_post)

    def handle_get(self):
        return _get_file_upload_form(f'{self.name} Model: convert PDF to data')

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        return [layout_document]

    def handle_post(self):  # pylint: disable=too-many-locals
        data = get_required_post_data()
        with TemporaryDirectory(suffix='-request') as temp_dir:
            temp_path = Path(temp_dir)
            pdf_path = temp_path / 'test.pdf'
            output_path = temp_path / 'test.lxml'
            first_page = get_int_request_arg(RequestArgs.FIRST_PAGE)
            last_page = get_int_request_arg(RequestArgs.LAST_PAGE)
            output_format = (
                request.args.get(RequestArgs.OUTPUT_FORMAT) or DEFAULT_MODEL_OUTPUT_FORMAT
            )
            assert output_format in VALID_MODEL_OUTPUT_FORMATS, \
                f'{output_format} not in {VALID_MODEL_OUTPUT_FORMATS}'
            pdf_path.write_bytes(data)
            self.pdfalto_wrapper.convert_pdf_to_pdfalto_xml(
                str(pdf_path),
                str(output_path),
                first_page=first_page,
                last_page=last_page
            )
            xml_content = output_path.read_bytes()
            root = etree.fromstring(xml_content)
            layout_document_iterable = self.iter_filter_layout_document(
                normalize_layout_document(
                    parse_alto_root(root)
                )
            )
            data_generator = self.model.get_data_generator(
                DocumentFeaturesContext(
                    app_features_context=self.app_features_context
                )
            )
            data_lines = data_generator.iter_data_lines_for_layout_documents(
                layout_document_iterable
            )
            response_type = 'text/plain'
            if output_format == ModelOutputFormats.RAW_DATA:
                response_content = '\n'.join(data_lines) + '\n'
            else:
                texts, features = load_data_crf_lines(data_lines)
                LOGGER.info('texts length: %d', len(texts))
                if not len(texts):  # pylint: disable=len-as-condition
                    tag_result = []
                else:
                    texts = texts.tolist()
                    tag_result = self.model.predict_labels(
                        texts=texts, features=features, output_format=None
                    )
                LOGGER.debug('tag_result: %s', tag_result)
                formatted_tag_result_iterable = iter_format_tag_result(
                    tag_result,
                    output_format=output_format,
                    expected_tag_result=None,
                    texts=texts,
                    features=features,
                    model_name=self.model_name
                )
                response_content = ''.join(formatted_tag_result_iterable)
                if output_format == TagOutputFormats.JSON:
                    response_type = 'application/json'
            LOGGER.debug('response_content: %r', response_content)
        headers = None
        return Response(response_content, headers=headers, mimetype=response_type)


class SegmentedModelNestedBluePrint(ModelNestedBluePrint):
    def __init__(
        self,
        *args,
        segmentation_model: Model,
        segmentation_labels: List[str],
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.segmentation_model = segmentation_model
        self.segmentation_labels = segmentation_labels

    def iter_filter_layout_document_by_segmentation_labels(
        self,
        layout_document: LayoutDocument,
        segmentation_labels: List[str]
    ) -> Iterable[LayoutDocument]:
        assert self.segmentation_model is not None
        segmentation_label_result = (
            self.segmentation_model.get_label_layout_document_result(
                layout_document,
                app_features_context=self.app_features_context
            )
        )
        for segmentation_label in segmentation_labels:
            layout_document = segmentation_label_result.get_filtered_document_by_label(
                segmentation_label
            ).remove_empty_blocks()
            if not layout_document:
                LOGGER.info(
                    'empty document for segmentation label %r, available labels: %r',
                    segmentation_label,
                    segmentation_label_result.get_available_labels()
                )
                continue
            yield layout_document

    def filter_layout_document_by_segmentation_label(
        self,
        layout_document: LayoutDocument,
        segmentation_label: str
    ) -> LayoutDocument:
        for filtered_layout_document in self.iter_filter_layout_document_by_segmentation_labels(
            layout_document,
            segmentation_labels=[segmentation_label]
        ):
            return filtered_layout_document
        return LayoutDocument(pages=[])

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        if get_bool_request_arg(RequestArgs.NO_USE_SEGMENTATION, default_value=False):
            return [layout_document]
        return self.iter_filter_layout_document_by_segmentation_labels(
            layout_document, segmentation_labels=self.segmentation_labels
        )


class NameHeaderModelNestedBluePrint(SegmentedModelNestedBluePrint):
    def __init__(
        self,
        *args,
        header_model: Model,
        merge_raw_authors: bool,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.header_model = header_model
        self.merge_raw_authors = merge_raw_authors

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        header_layout_document = self.filter_layout_document_by_segmentation_label(
            layout_document, '<header>'
        )
        labeled_layout_tokens = self.header_model.predict_labels_for_layout_document(
            header_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        semantic_raw_authors_list = list(
            SemanticMixedContentWrapper(list(
                self.header_model.iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawAuthors)
        )
        LOGGER.info('semantic_raw_authors_list count: %d', len(semantic_raw_authors_list))
        LOGGER.info('merge_raw_authors: %s', self.merge_raw_authors)
        if self.merge_raw_authors:
            return [
                LayoutDocument.for_blocks([
                    block
                    for semantic_raw_authors in semantic_raw_authors_list
                    for block in semantic_raw_authors.iter_blocks()
                ]).remove_empty_blocks()
            ]
        return [
            LayoutDocument.for_blocks(
                list(semantic_raw_authors.iter_blocks())
            ).remove_empty_blocks()
            for semantic_raw_authors in semantic_raw_authors_list
        ]


class AffiliationAddressModelNestedBluePrint(SegmentedModelNestedBluePrint):
    def __init__(self, *args, header_model: Model, **kwargs):
        super().__init__(*args, **kwargs)
        self.header_model = header_model

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        header_layout_document = self.filter_layout_document_by_segmentation_label(
            layout_document, '<header>'
        )
        labeled_layout_tokens = self.header_model.predict_labels_for_layout_document(
            header_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        semantic_raw_aff_address_list = list(
            SemanticMixedContentWrapper(list(
                self.header_model.iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawAffiliationAddress)
        )
        LOGGER.info('semantic_raw_aff_address_list count: %d', len(semantic_raw_aff_address_list))
        return [
            LayoutDocument.for_blocks(
                list(semantic_raw_aff_address.iter_blocks())
            ).remove_empty_blocks()
            for semantic_raw_aff_address in semantic_raw_aff_address_list
        ]


class CitationModelNestedBluePrint(SegmentedModelNestedBluePrint):
    def __init__(self, *args, reference_segmenter_model: Model, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_segmenter_model = reference_segmenter_model

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        references_layout_document = self.filter_layout_document_by_segmentation_label(
            layout_document, '<references>'
        )
        labeled_layout_tokens = self.reference_segmenter_model.predict_labels_for_layout_document(
            references_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        semantic_raw_references = list(
            SemanticMixedContentWrapper(list(
                self.reference_segmenter_model.iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawReference)
        )
        LOGGER.info('semantic_raw_references count: %d', len(semantic_raw_references))
        return [
            LayoutDocument.for_blocks(
                [semantic_raw_reference.view_by_type(SemanticRawReferenceText).merged_block]
            ).remove_empty_blocks()
            for semantic_raw_reference in semantic_raw_references
        ]


class NameCitationModelNestedBluePrint(SegmentedModelNestedBluePrint):
    def __init__(
        self,
        *args,
        reference_segmenter_model: Model,
        citation_model: Model,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.reference_segmenter_model = reference_segmenter_model
        self.citation_model = citation_model

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        references_layout_document = self.filter_layout_document_by_segmentation_label(
            layout_document, '<references>'
        )
        labeled_layout_tokens = self.reference_segmenter_model.predict_labels_for_layout_document(
            references_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        semantic_raw_references = list(
            SemanticMixedContentWrapper(list(
                self.reference_segmenter_model.iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawReference)
        )
        LOGGER.info('semantic_raw_references count: %d', len(semantic_raw_references))
        raw_reference_documents = [
            LayoutDocument.for_blocks(
                [semantic_raw_reference.view_by_type(SemanticRawReferenceText).merged_block]
            ).remove_empty_blocks()
            for semantic_raw_reference in semantic_raw_references
        ]
        citation_labeled_layout_tokens_list = (
            self.citation_model.predict_labels_for_layout_documents(
                raw_reference_documents,
                app_features_context=self.app_features_context
            )
        )
        raw_authors = [
            raw_author
            for citation_labeled_layout_tokens in citation_labeled_layout_tokens_list
            for ref in (
                self.citation_model.iter_semantic_content_for_labeled_layout_tokens(
                    citation_labeled_layout_tokens
                )
            )
            if isinstance(ref, SemanticReference)
            for raw_author in ref.iter_by_type(SemanticRawAuthors)
        ]
        return [
            LayoutDocument.for_blocks([raw_author.merged_block]).remove_empty_blocks()
            for raw_author in raw_authors
        ]


class FullTextChildModelNestedBluePrint(SegmentedModelNestedBluePrint):
    def __init__(
        self,
        *args,
        fulltext_model: Model,
        semantic_type: Type[T_SemanticContentWrapper],
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.fulltext_model = fulltext_model
        self.semantic_type = semantic_type

    def iter_filter_layout_document(
        self, layout_document: LayoutDocument
    ) -> Iterable[LayoutDocument]:
        fulltext_layout_documents = list(self.iter_filter_layout_document_by_segmentation_labels(
            layout_document, self.segmentation_labels
        ))
        fulltext_labeled_layout_tokens_list = (
            self.fulltext_model.predict_labels_for_layout_documents(
                fulltext_layout_documents,
                app_features_context=self.app_features_context
            )
        )
        LOGGER.debug('fulltext_labeled_layout_tokens_list: %r', fulltext_labeled_layout_tokens_list)
        semanti_content_list = [
            semantic_content
            for fulltext_labeled_layout_tokens in fulltext_labeled_layout_tokens_list
            for semantic_content in iter_by_semantic_type_recursively(
                self.fulltext_model.iter_semantic_content_for_labeled_layout_tokens(
                    fulltext_labeled_layout_tokens
                ),
                self.semantic_type
            )
        ]
        LOGGER.debug('semanti_content_list: %s', semanti_content_list)
        return [
            LayoutDocument.for_blocks([semanti_content.merged_block]).remove_empty_blocks()
            for semanti_content in semanti_content_list
        ]


class FigureModelNestedBluePrint(FullTextChildModelNestedBluePrint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, semantic_type=SemanticRawFigure, **kwargs)


class TableModelNestedBluePrint(FullTextChildModelNestedBluePrint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, semantic_type=SemanticRawTable, **kwargs)


class ApiBlueprint(Blueprint):
    def __init__(
        self,
        sciencebeam_parser: ScienceBeamParser
    ):
        super().__init__('api', __name__)
        LOGGER.debug('sciencebeam_parser: %r', sciencebeam_parser)
        self.sciencebeam_parser = sciencebeam_parser
        self.route('/')(self.api_root)
        self.route("/pdfalto", methods=['GET'])(self.pdfalto_form)
        self.route("/pdfalto", methods=['POST'])(self.pdfalto)
        self.route("/processHeaderDocument", methods=['GET'])(
            self.process_header_document_api_form
        )
        self.route("/processHeaderDocument", methods=['POST'])(
            self.process_header_document_api
        )
        self.route("/processFulltextDocument", methods=['GET'])(
            self.process_fulltext_document_api_form
        )
        self.route("/processFulltextDocument", methods=['POST'])(
            self.process_fulltext_document_api
        )
        self.route("/processReferences", methods=['GET'])(
            self.process_references_api_form
        )
        self.route("/processReferences", methods=['POST'])(
            self.process_references_api
        )
        self.route("/processFulltextAssetDocument", methods=['GET'])(
            self.process_pdf_to_tei_assets_zip_form
        )
        self.route("/processFulltextAssetDocument", methods=['POST'])(
            self.process_pdf_to_tei_assets_zip
        )
        self.route("/convert", methods=['GET'])(self.process_convert_api_form)
        self.route("/convert", methods=['POST'])(self.process_convert_api)
        self.pdfalto_wrapper = sciencebeam_parser.pdfalto_wrapper
        self.app_context = sciencebeam_parser.app_context
        self.fulltext_processor_config = sciencebeam_parser.fulltext_processor_config
        self.fulltext_models = sciencebeam_parser.fulltext_models
        self.app_features_context = sciencebeam_parser.app_features_context
        fulltext_models = self.fulltext_models
        app_features_context = self.app_features_context
        ModelNestedBluePrint(
            'Segmentation',
            model=fulltext_models.segmentation_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context
        ).add_routes(self, '/models/segmentation')
        SegmentedModelNestedBluePrint(
            'Header',
            model=fulltext_models.header_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<header>']
        ).add_routes(self, '/models/header')
        NameHeaderModelNestedBluePrint(
            'Name Header',
            model=fulltext_models.name_header_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<header>'],
            header_model=fulltext_models.header_model,
            merge_raw_authors=self.fulltext_processor_config.merge_raw_authors
        ).add_routes(self, '/models/name-header')
        AffiliationAddressModelNestedBluePrint(
            'Affiliation Address',
            model=fulltext_models.affiliation_address_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<header>'],
            header_model=fulltext_models.header_model
        ).add_routes(self, '/models/affiliation-address')
        fulltext_segmentation_labels = ['<body>', '<acknowledgement>', '<annex>']
        SegmentedModelNestedBluePrint(
            'FullText',
            model=fulltext_models.fulltext_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=fulltext_segmentation_labels
        ).add_routes(self, '/models/fulltext')
        FigureModelNestedBluePrint(
            'Figure',
            model=fulltext_models.figure_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=fulltext_segmentation_labels,
            fulltext_model=fulltext_models.fulltext_model
        ).add_routes(self, '/models/figure')
        TableModelNestedBluePrint(
            'Table',
            model=fulltext_models.table_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=fulltext_segmentation_labels,
            fulltext_model=fulltext_models.fulltext_model
        ).add_routes(self, '/models/table')
        SegmentedModelNestedBluePrint(
            'Reference Segmenter',
            model=fulltext_models.reference_segmenter_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<references>']
        ).add_routes(self, '/models/reference-segmenter')
        CitationModelNestedBluePrint(
            'Citation (Reference)',
            model=fulltext_models.citation_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<references>'],
            reference_segmenter_model=fulltext_models.reference_segmenter_model
        ).add_routes(self, '/models/citation')
        NameCitationModelNestedBluePrint(
            'Name Citaton',
            model=fulltext_models.name_citation_model,
            pdfalto_wrapper=self.pdfalto_wrapper,
            app_features_context=app_features_context,
            segmentation_model=fulltext_models.segmentation_model,
            segmentation_labels=['<references>'],
            reference_segmenter_model=fulltext_models.reference_segmenter_model,
            citation_model=fulltext_models.citation_model
        ).add_routes(self, '/models/name-citation')

    def api_root(self):
        return jsonify({
            'links': {
                'pdfalto': url_for('.pdfalto')
            }
        })

    def pdfalto_form(self):
        return _get_file_upload_form('PdfAlto convert PDF to LXML')

    @contextmanager
    def _get_new_sciencebeam_parser_session(
        self, **kwargs
    ) -> Iterator[ScienceBeamParserSession]:
        with self.sciencebeam_parser.get_new_session(**kwargs) as session:
            first_page = get_int_request_arg(RequestArgs.FIRST_PAGE)
            last_page = get_int_request_arg(RequestArgs.LAST_PAGE)
            session.document_request_parameters.first_page = first_page
            session.document_request_parameters.last_page = last_page
            yield session

    @contextmanager
    def _get_new_sciencebeam_parser_source(
        self, **kwargs
    ) -> Iterator[ScienceBeamParserSessionSource]:
        data_wrapper = get_data_wrapper_with_improved_media_type_or_filename(
            get_required_post_data_wrapper()
        )
        with self._get_new_sciencebeam_parser_session(**kwargs) as session:
            source_path = session.temp_path / 'source.file'
            source_path.write_bytes(data_wrapper.data)
            yield session.get_source(
                source_path=str(source_path),
                source_media_type=data_wrapper.media_type
            )

    def pdfalto(self):
        response_type = 'text/xml'
        with self._get_new_sciencebeam_parser_source() as source:
            return send_file(
                source.get_local_file_for_response_media_type(
                    MediaTypes.ALTO_XML
                ),
                mimetype=response_type
            )

    def _process_pdf_to_response_media_type(
        self,
        response_media_type: str,
        fulltext_processor_config: FullTextProcessorConfig
    ):
        try:
            LOGGER.debug('creating session source')
            with self._get_new_sciencebeam_parser_source(
                fulltext_processor_config=fulltext_processor_config
            ) as source:
                LOGGER.debug('created session source: %r', source)
                actual_response_media_type = response_media_type
                if response_media_type in {
                    MediaTypes.TEI_XML, MediaTypes.JATS_XML
                }:
                    actual_response_media_type = MediaTypes.XML
                elif response_media_type in {
                    MediaTypes.TEI_ZIP, MediaTypes.JATS_ZIP
                }:
                    actual_response_media_type = MediaTypes.ZIP
                result_file = source.get_local_file_for_response_media_type(
                    response_media_type
                )
                LOGGER.debug('result_file: %r', result_file)
                assert isinstance(result_file, str)
                return send_file(
                    result_file,
                    mimetype=actual_response_media_type
                )
        except BadRequestScienceBeamParserError as exc:
            LOGGER.warning('bad request: %r', exc)
            return BadRequest(str(exc))

    def process_header_document_api_form(self):
        return _get_file_upload_form('Convert PDF to TEI (Header)')

    def process_header_document_api(self):
        response_media_type = assert_and_get_first_accept_matching_media_type(
            [MediaTypes.TEI_XML, MediaTypes.JATS_XML]
        )
        return self._process_pdf_to_response_media_type(
            response_media_type,
            fulltext_processor_config=self.fulltext_processor_config.get_for_header_document()
        )

    def process_fulltext_document_api_form(self):
        return _get_file_upload_form('Convert PDF to TEI (Full Text)')

    def process_fulltext_document_api(self):
        response_media_type = assert_and_get_first_accept_matching_media_type(
            [MediaTypes.TEI_XML, MediaTypes.JATS_XML]
        )
        return self._process_pdf_to_response_media_type(
            response_media_type,
            fulltext_processor_config=self.fulltext_processor_config
        )

    def process_references_api_form(self):
        return _get_file_upload_form('Convert PDF to TEI (References)')

    def process_references_api(self):
        response_media_type = assert_and_get_first_accept_matching_media_type(
            [MediaTypes.TEI_XML, MediaTypes.JATS_XML]
        )
        return self._process_pdf_to_response_media_type(
            response_media_type,
            fulltext_processor_config=self.fulltext_processor_config.get_for_requested_field_names({
                RequestFieldNames.REFERENCES
            })
        )

    def process_pdf_to_tei_assets_zip_form(self):
        return _get_file_upload_form('Convert PDF to TEI Assets ZIP')

    def process_pdf_to_tei_assets_zip(self):
        response_media_type = assert_and_get_first_accept_matching_media_type(
            [MediaTypes.TEI_ZIP, MediaTypes.JATS_ZIP]
        )
        return self._process_pdf_to_response_media_type(
            response_media_type,
            fulltext_processor_config=self.fulltext_processor_config
        )

    def process_convert_api_form(self):
        return _get_file_upload_form('Convert API')

    def process_convert_api(self):
        response_media_type = assert_and_get_first_accept_matching_media_type(
            [
                MediaTypes.JATS_XML, MediaTypes.TEI_XML,
                MediaTypes.JATS_ZIP, MediaTypes.TEI_ZIP,
                MediaTypes.PDF
            ]
        )
        includes_list = parse_comma_separated_value(
            get_str_request_arg(RequestArgs.INCLUDES) or ''
        )
        LOGGER.info('includes_list: %r', includes_list)
        fulltext_processor_config = self.fulltext_processor_config.get_for_requested_field_names(
            set(includes_list)
        )
        LOGGER.info('fulltext_processor_config: %r', fulltext_processor_config)
        return self._process_pdf_to_response_media_type(
            response_media_type,
            fulltext_processor_config=fulltext_processor_config
        )



import argparse
import logging
import os
from typing import Iterable, List, Optional, Sequence, Tuple

from lxml import etree

from sciencebeam_trainer_delft.utils.io import (
    auto_download_input_file
)
from sciencebeam_trainer_delft.sequence_labelling.reader import (
    load_data_crf_lines
)
from sciencebeam_trainer_delft.sequence_labelling.tag_formatter import (
    TagOutputFormats,
    iter_format_tag_result
)

from sciencebeam_parser.utils.io import (
    auto_uploading_output_file,
    glob
)

from sciencebeam_parser.document.layout_document import (
    LayoutBlock,
    LayoutDocument
)

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext,
    LabeledLayoutToken,
    ModelDataGenerator
)
from sciencebeam_parser.models.training_data import TrainingTeiParser

from sciencebeam_parser.resources.default_config import DEFAULT_CONFIG_FILE
from sciencebeam_parser.config.config import AppConfig
from sciencebeam_parser.app.parser import ScienceBeamParser


LOGGER = logging.getLogger(__name__)


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        'ScienceBeam Parser: Generate DELFT Training Data'
    )
    parser.add_argument(
        '--model-name',
        type=str,
        required=True
    )
    parser.add_argument(
        '--tei-source-path',
        type=str,
        required=True
    )
    parser.add_argument(
        '--raw-source-path',
        type=str,
        required=False
    )
    parser.add_argument(
        '--delft-output-path',
        type=str,
        required=True
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help='Enable debug logging'
    )
    return parser.parse_args(argv)


def translate_tags_IOB_to_grobid(tag: str) -> str:
    """
    Convert labels from IOB2 to the ones used by GROBID (expected by the wapiti model)
    """
    if tag == 'O':
        # outside
        return '<other>'
    if tag.startswith('B-'):
        # begin
        return 'I-' + tag[2:]
    if tag.startswith('I-'):
        # inside
        return '' + tag[2:]
    return tag


def translate_tag_result_tags_IOB_to_grobid(
    tag_result: Sequence[Sequence[Tuple[str, str]]]
) -> List[List[Tuple[str, str]]]:
    return [
        [
            (token_text, translate_tags_IOB_to_grobid(tag))
            for token_text, tag in doc_tag_result
        ]
        for doc_tag_result in tag_result
    ]


def get_tag_result_for_labeled_layout_tokens_list(
    labeled_layout_tokens_list: Sequence[Sequence[LabeledLayoutToken]]
) -> List[List[Tuple[str, str]]]:
    return [
        [
            (
                labeled_layout_token.layout_token.text,
                labeled_layout_token.label
            )
            for labeled_layout_token in labeled_layout_tokens
        ]
        for labeled_layout_tokens in labeled_layout_tokens_list
    ]


def get_raw_file_for_tei_file(
    tei_file: str,
    raw_source_path: str
) -> str:
    compression_suffix = ''
    if tei_file.endswith('.gz'):
        compression_suffix = '.gz'
        tei_file = tei_file[:-len(compression_suffix)]
    tei_suffix = '.tei.xml'
    assert tei_file.endswith(tei_suffix)
    return os.path.join(
        raw_source_path,
        os.path.basename(tei_file[:-len(tei_suffix)] + compression_suffix)
    )


def get_raw_file_list_for_tei_file_list(
    tei_file_list: Iterable[str],
    raw_source_path: str
) -> Sequence[str]:
    return [
        get_raw_file_for_tei_file(tei_file, raw_source_path=raw_source_path)
        for tei_file in tei_file_list
    ]


def get_training_tei_parser_for_model_name(
    model_name: str,
    sciencebeam_parser: ScienceBeamParser
) -> TrainingTeiParser:
    model = sciencebeam_parser.fulltext_models.get_sequence_model_by_name(model_name)
    try:
        training_tei_parser = model.get_training_tei_parser()
        assert training_tei_parser is not None
        return training_tei_parser
    except NotImplementedError as exc:
        training_tei_parser = None
        raise RuntimeError('unsupported model: %r' % model_name) from exc


def get_data_generator_for_model_name(
    model_name: str,
    sciencebeam_parser: ScienceBeamParser
) -> ModelDataGenerator:
    model = sciencebeam_parser.fulltext_models.get_sequence_model_by_name(model_name)
    return model.get_data_generator(
        document_features_context=DocumentFeaturesContext(
            app_features_context=sciencebeam_parser.app_features_context
        )
    )


def iter_generate_delft_training_data_lines_for_document(  # pylint: disable=too-many-locals
    tei_file: str,
    raw_file: Optional[str],
    training_tei_parser: TrainingTeiParser,
    data_generator: ModelDataGenerator
) -> Iterable[str]:
    with auto_download_input_file(
        tei_file,
        auto_decompress=True
    ) as local_tei_file:
        tei_root = etree.parse(local_tei_file).getroot()
    labeled_layout_tokens_list = (
        training_tei_parser.parse_training_tei_to_labeled_layout_tokens_list(
            tei_root
        )
    )
    LOGGER.debug('labeled_layout_tokens_list: %r', labeled_layout_tokens_list)
    translated_tag_result = translate_tag_result_tags_IOB_to_grobid(
        get_tag_result_for_labeled_layout_tokens_list(
            labeled_layout_tokens_list
        )
    )
    LOGGER.debug('translated_tag_result: %r', translated_tag_result)
    if raw_file:
        with auto_download_input_file(
            raw_file,
            auto_decompress=True
        ) as local_raw_file:
            with open(local_raw_file, 'r', encoding='utf-8') as raw_fp:
                texts, features = load_data_crf_lines(
                    raw_fp
                )
        assert len(texts) == len(translated_tag_result)
        for doc_tokens, doc_tag_result in zip(texts, translated_tag_result):
            assert len(doc_tokens) == len(doc_tag_result)
    else:
        layout_documents = [
            LayoutDocument.for_blocks([
                LayoutBlock.for_tokens([
                    labeled_layout_token.layout_token
                    for labeled_layout_token in labeled_layout_tokens
                ])
            ])
            for labeled_layout_tokens in labeled_layout_tokens_list
        ]
        LOGGER.debug('layout_documents: %r', layout_documents)
        data_line_iterable = list(data_generator.iter_data_lines_for_layout_documents(
            layout_documents
        ))
        _texts, features = load_data_crf_lines(data_line_iterable)
    LOGGER.debug('features: %r', features)
    yield from iter_format_tag_result(
        tag_result=translated_tag_result,
        output_format=TagOutputFormats.DATA,
        texts=None,
        features=features
    )


def generate_delft_training_data(
    model_name: str,
    tei_source_path: str,
    raw_source_path: str,
    delft_output_path: str,
    sciencebeam_parser: ScienceBeamParser
):
    training_tei_parser = get_training_tei_parser_for_model_name(
        model_name,
        sciencebeam_parser=sciencebeam_parser
    )
    data_generator = get_data_generator_for_model_name(
        model_name,
        sciencebeam_parser=sciencebeam_parser
    )
    LOGGER.debug('tei_source_path: %r', tei_source_path)
    tei_file_list = glob(tei_source_path)
    if not tei_file_list:
        raise RuntimeError('no files found for file pattern %r' % tei_source_path)
    LOGGER.info('tei_file_list: %r', tei_file_list)
    if raw_source_path:
        raw_file_list: Sequence[Optional[str]] = get_raw_file_list_for_tei_file_list(
            tei_file_list,
            raw_source_path=raw_source_path
        )
    else:
        raw_file_list = [None] * len(tei_file_list)
    LOGGER.info('raw_file_list: %r', raw_file_list)
    LOGGER.info('writing to : %r', delft_output_path)
    with auto_uploading_output_file(
        delft_output_path,
        mode='w',
        encoding='utf-8',
    ) as data_fp:
        for document_index, (tei_file, raw_file) in enumerate(zip(tei_file_list, raw_file_list)):
            if document_index > 0:
                data_fp.write('\n\n')
            data_fp.writelines(iter_generate_delft_training_data_lines_for_document(
                tei_file=tei_file,
                raw_file=raw_file,
                training_tei_parser=training_tei_parser,
                data_generator=data_generator
            ))


def run(args: argparse.Namespace):
    LOGGER.info('args: %r', args)
    config = AppConfig.load_yaml(
        DEFAULT_CONFIG_FILE
    )
    sciencebeam_parser = ScienceBeamParser.from_config(config)
    generate_delft_training_data(
        model_name=args.model_name,
        tei_source_path=args.tei_source_path,
        raw_source_path=args.raw_source_path,
        delft_output_path=args.delft_output_path,
        sciencebeam_parser=sciencebeam_parser
    )


def main(argv: Optional[List[str]] = None):
    LOGGER.debug('argv: %r', argv)
    args = parse_args(argv)
    if args.debug:
        for name in [__name__, 'sciencebeam_parser', 'sciencebeam_trainer_delft']:
            logging.getLogger(name).setLevel('DEBUG')
    run(args)


if __name__ == '__main__':
    logging.basicConfig(level='INFO')

    main()



import re
from typing import Sequence

# mostly copied from:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/utilities/TextUtilities.java#L773-L948
# added: '\u2010' .. '\u2015', '`'
REPLACEMENT_CHARACTER_BY_CHARACTER_MAP = {
    '`': "'",
    '\uFB00': 'ff',
    '\uFB01': 'fi',
    '\uFB02': 'fl',
    '\uFB03': 'ffi',
    '\uFB04': 'ffl',
    '\uFB05': 'ft',
    '\uFB06': 'st',
    '\u00E6': 'ae',
    '\u00C6': 'AE',
    '\u0153': 'oe',
    '\u0152': 'OE',
    '\u2010': '-',
    '\u2011': '-',
    '\u2012': '-',
    '\u2013': '-',
    '\u2014': '-',
    '\u2015': '-',
    '\u201C': '"',
    '\u201D': '"',
    '\u201E': '"',
    '\u201F': '"',
    '\u2019': '\'',
    '\u2018': '\'',
    '\u2022': '•',
    '\u2023': '•',
    '\u2043': '•',
    '\u204C': '•',
    '\u204D': '•',
    '\u2219': '•',
    '\u25C9': '•',
    '\u25D8': '•',
    '\u25E6': '•',
    '\u2619': '•',
    '\u2765': '•',
    '\u2767': '•',
    '\u29BE': '•',
    '\u29BF': '•',
    '\u2217': '*'
}


REPLACEMENT_CHARACTER_BY_CHARACTER_TRANSLATION = str.maketrans(
    REPLACEMENT_CHARACTER_BY_CHARACTER_MAP
)


def normalize_text(text: str) -> str:
    return re.sub(
        r'\s{2,}',
        ' ',
        re.sub(
            r'\s*\n\s*',
            '\n',
            text.translate(REPLACEMENT_CHARACTER_BY_CHARACTER_TRANSLATION),
            flags=re.RegexFlag.MULTILINE
        )
    )


def remove_whitespace(text: str) -> str:
    return re.sub(r'\s', '', text)


def parse_comma_separated_value(s: str, sep: str = ',') -> Sequence[str]:
    s = s.strip()
    if not s:
        return []
    return [item.strip() for item in s.split(sep)]



from typing import NamedTuple, Optional


from sciencebeam_parser.utils.media_types import (
    MediaTypes,
    guess_extension_for_media_type,
    guess_media_type_for_filename
)


DEFAULT_FILENAME = 'file'


class MediaDataWrapper(NamedTuple):
    data: bytes
    media_type: str
    filename: Optional[str] = None


def get_data_wrapper_with_improved_media_type_or_filename(
    data_wrapper: MediaDataWrapper
) -> MediaDataWrapper:
    if not data_wrapper.filename:
        return data_wrapper._replace(filename='%s%s' % (
            DEFAULT_FILENAME, guess_extension_for_media_type(data_wrapper.media_type) or ''
        ))
    if not data_wrapper.media_type or data_wrapper.media_type == MediaTypes.OCTET_STREAM:
        media_type = guess_media_type_for_filename(data_wrapper.filename)
        if media_type:
            return data_wrapper._replace(media_type=media_type)
    return data_wrapper



import logging
import re
from typing import Iterable, Mapping, Optional, Tuple

from sciencebeam_parser.document.semantic_document import (
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticFigureCitation,
    SemanticHeading,
    SemanticLabel,
    SemanticNote,
    SemanticParagraph,
    SemanticRawEquation,
    SemanticRawEquationContent,
    SemanticRawFigure,
    SemanticRawTable,
    SemanticReferenceCitation,
    SemanticSection,
    SemanticSectionTypes,
    SemanticTableCitation,
    SemanticTitle
)
from sciencebeam_parser.document.layout_document import LayoutBlock, LayoutTokensText
from sciencebeam_parser.models.extract import SimpleModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<figure>': SemanticRawFigure,
    '<table>': SemanticRawTable
}


PARAGRAPH_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<figure_marker>': SemanticFigureCitation,
    '<table_marker>': SemanticTableCitation,
    '<citation_marker>': SemanticReferenceCitation
}


HEADER_LABEL_REGEX = r'(\d+\.?(?:\d+\.?)*)\s*(\D.*)'


def get_section_label_and_title_from_layout_block(
    layout_block: LayoutBlock
) -> Tuple[Optional[LayoutBlock], LayoutBlock]:
    if not layout_block:
        return None, layout_block
    layout_tokens_text = LayoutTokensText(layout_block)
    text = str(layout_tokens_text)
    m = re.match(HEADER_LABEL_REGEX, text, re.IGNORECASE)
    if not m:
        return None, layout_block
    label_end = m.end(1)
    title_start = m.start(2)
    LOGGER.debug('label_end: %d, title_start: %d (text: %r)', label_end, title_start, text)
    section_label_layout_block = LayoutBlock.for_tokens(list(
        layout_tokens_text.iter_layout_tokens_between(0, label_end)
    ))
    section_title_layout_block = LayoutBlock.for_tokens(list(
        layout_tokens_text.iter_layout_tokens_between(title_start, len(text))
    ))
    return section_label_layout_block, section_title_layout_block


class FullTextSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def add_paragraph_content(
        self,
        paragraph: SemanticParagraph,
        name: str,
        layout_block: LayoutBlock
    ):
        semantic_content_class = PARAGRAPH_SEMANTIC_CONTENT_CLASS_BY_TAG.get(name)
        if semantic_content_class:
            paragraph.add_content(semantic_content_class(layout_block=layout_block))
            return
        paragraph.add_block_content(layout_block)

    def get_semantic_heading(self, layout_block: LayoutBlock):
        section_label_layout_block, section_title_layout_block = (
            get_section_label_and_title_from_layout_block(layout_block)
        )
        if section_label_layout_block:
            return SemanticHeading([
                SemanticLabel(layout_block=section_label_layout_block),
                SemanticTitle(layout_block=section_title_layout_block)
            ])
        return SemanticHeading([
            SemanticTitle(layout_block=section_title_layout_block)
        ])

    def get_raw_equation_child_semantic_content(
        self,
        name: str,
        layout_block: LayoutBlock
    ):
        if name == '<equation_label>':
            return SemanticLabel(layout_block=layout_block)
        if name == '<equation>':
            return SemanticRawEquationContent(layout_block=layout_block)
        return self.get_semantic_content_for_entity_name(
            name, layout_block=layout_block
        )

    def iter_semantic_content_for_entity_blocks(  # noqa pylint: disable=arguments-differ, too-many-branches
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        section_type: str = SemanticSectionTypes.OTHER,
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        section: Optional[SemanticSection] = None
        paragraph: Optional[SemanticParagraph] = None
        raw_equation: Optional[SemanticRawEquation] = None
        _previous_tag: Optional[str] = None
        for name, layout_block in entity_tokens:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug('entity_block: %r, %r', name, layout_block.text)
            previous_tag = _previous_tag
            _previous_tag = name
            if name in {'O'}:
                LOGGER.debug('ignoring content (%r): %r', name, layout_block)
                note_type = 'fulltext:other' if name == 'O' else name
                if section:
                    section.add_note(layout_block, note_type=note_type)
                else:
                    yield SemanticNote(
                        layout_block=layout_block,
                        note_type=note_type
                    )
                continue
            if name == '<section>':
                paragraph = None
                raw_equation = None
                if section:
                    yield section
                section = SemanticSection(section_type=section_type)
                section.add_content(self.get_semantic_heading(layout_block))
                continue
            if not section:
                section = SemanticSection(section_type=section_type)
            if name in SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG:
                section.add_content(self.get_semantic_content_for_entity_name(
                    name, layout_block=layout_block
                ))
                continue
            # treat everything else as paragraph content
            if (
                not paragraph
                or (
                    name == '<paragraph>'
                    and previous_tag == '<paragraph>'
                )
            ):
                paragraph = section.add_new_paragraph()
            if name in {'<equation>', '<equation_label>'}:
                semantic_content = self.get_raw_equation_child_semantic_content(
                    name, layout_block=layout_block
                )
                if (
                    isinstance(semantic_content, SemanticRawEquationContent)
                    and raw_equation
                    and raw_equation.has_type(SemanticRawEquationContent)
                ):
                    LOGGER.debug('already has equation content, start new one')
                    raw_equation = None
                if not raw_equation:
                    raw_equation = SemanticRawEquation()
                    paragraph.add_content(raw_equation)
                raw_equation.add_content(semantic_content)
                continue
            raw_equation = None
            self.add_paragraph_content(
                paragraph, name, layout_block
            )
        if section:
            yield section



import logging

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticGraphic,
)
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    format_coordinates
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class GraphicTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticGraphic)
        semantic_graphic = semantic_content
        layout_graphic = semantic_graphic.layout_graphic
        children = [
            context.get_default_attributes_for_semantic_content(
                semantic_graphic,
                enable_coordinates=False
            )
        ]
        if semantic_graphic.relative_path:
            children.append({'url': semantic_graphic.relative_path})
        if layout_graphic:
            if layout_graphic.coordinates:
                children.append({'coords': format_coordinates(layout_graphic.coordinates)})
            if layout_graphic.graphic_type:
                children.append({'type': layout_graphic.graphic_type})
        return TEI_E('graphic', *children)



from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Set


class TextLookUp(ABC):
    @abstractmethod
    def contains(self, text: str) -> bool:
        pass


class SimpleTextLookUp(TextLookUp):
    def __init__(self, valid_texts: Set[str]):
        self.valid_texts = {s.lower() for s in valid_texts}

    def contains(self, text: str) -> bool:
        return text.lower() in self.valid_texts


class MergedTextLookUp(TextLookUp):
    def __init__(self, lookups: Sequence[Optional[TextLookUp]]):
        self.lookups: List[TextLookUp] = [
            lookup
            for lookup in lookups
            if lookup is not None
        ]

    def contains(self, text: str) -> bool:
        for lookup in self.lookups:
            if lookup.contains(text):
                return True
        return False



import logging

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.segmentation.data import SegmentationDataGenerator
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.segmentation.training_data import (
    SegmentationTeiTrainingDataGenerator,
    SegmentationTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


class SegmentationModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> SegmentationDataGenerator:
        return SegmentationDataGenerator(
            document_features_context=document_features_context,
            use_first_token_of_block=(
                self.model_config.get('use_first_token_of_block', True)
            )
        )

    def get_tei_training_data_generator(self) -> SegmentationTeiTrainingDataGenerator:
        return SegmentationTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> SegmentationTrainingTeiParser:
        return SegmentationTrainingTeiParser()



import logging
import os
from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from time import monotonic
from typing import List, Optional, Set
from zipfile import ZipFile

from lxml import etree

from sciencebeam_trainer_delft.utils.download_manager import DownloadManager

from sciencebeam_parser.app.context import AppContext
from sciencebeam_parser.config.config import AppConfig, get_download_dir
from sciencebeam_parser.external.pdfalto.wrapper import PdfAltoWrapper
from sciencebeam_parser.external.pdfalto.parser import parse_alto_root
from sciencebeam_parser.external.wapiti.wrapper import LazyWapitiBinaryWrapper
from sciencebeam_parser.lookup.loader import load_lookup_from_config
from sciencebeam_parser.models.data import AppFeaturesContext
from sciencebeam_parser.document.layout_document import LayoutDocument
from sciencebeam_parser.document.semantic_document import (
    SemanticDocument,
    SemanticGraphic
)
from sciencebeam_parser.document.tei_document import get_tei_for_semantic_document
from sciencebeam_parser.processors.fulltext.models import FullTextModels
from sciencebeam_parser.resources.xslt import TEI_TO_JATS_XSLT_FILE
from sciencebeam_parser.transformers.doc_converter_wrapper import DocConverterWrapper
from sciencebeam_parser.transformers.xslt import XsltTransformerWrapper
from sciencebeam_parser.utils.lazy import LazyLoaded
from sciencebeam_parser.utils.media_types import (
    MediaTypes
)
from sciencebeam_parser.utils.text import normalize_text
from sciencebeam_parser.utils.tokenizer import get_tokenized_tokens
from sciencebeam_parser.processors.fulltext.api import (
    FullTextProcessor,
    FullTextProcessorConfig,
    FullTextProcessorDocumentContext,
    load_models
)


LOGGER = logging.getLogger(__name__)


TEMP_ALTO_XML_FILENAME = 'temp.lxml'


DOC_TO_PDF_SUPPORTED_MEDIA_TYPES = {
    MediaTypes.DOCX,
    MediaTypes.DOTX,
    MediaTypes.DOC,
    MediaTypes.RTF
}


JATS_MEDIA_TYPES = {MediaTypes.JATS_XML, MediaTypes.JATS_ZIP}
ASSET_ZIP_MEDIA_TYPES = {MediaTypes.TEI_ZIP, MediaTypes.JATS_ZIP}


def normalize_and_tokenize_text(text: str) -> List[str]:
    return get_tokenized_tokens(
        normalize_text(text),
        keep_whitespace=True
    )


def normalize_layout_document(
    layout_document: LayoutDocument,
    **kwargs
) -> LayoutDocument:
    return (
        layout_document
        .retokenize(tokenize_fn=normalize_and_tokenize_text)
        .remove_empty_blocks(**kwargs)
    )


def load_app_features_context(
    config: AppConfig,
    download_manager: DownloadManager
):
    return AppFeaturesContext(
        country_lookup=load_lookup_from_config(
            config.get('lookup', {}).get('country'),
            download_manager=download_manager
        ),
        first_name_lookup=load_lookup_from_config(
            config.get('lookup', {}).get('first_name'),
            download_manager=download_manager
        ),
        last_name_lookup=load_lookup_from_config(
            config.get('lookup', {}).get('last_name'),
            download_manager=download_manager
        )
    )


def create_asset_zip_for_semantic_document(
    zip_filename: str,
    semantic_document: SemanticDocument,
    relative_xml_filename: str,
    local_xml_filename: str
):
    semantic_graphic_list = list(semantic_document.iter_by_type_recursively(
        SemanticGraphic
    ))
    LOGGER.debug('semantic_graphic_list: %r', semantic_graphic_list)
    with ZipFile(zip_filename, 'w') as zip_file:
        zip_file.write(
            local_xml_filename,
            relative_xml_filename
        )
        for semantic_graphic in semantic_graphic_list:
            assert semantic_graphic.relative_path, \
                "graphic relative_path missing, ensure extract_graphic_assets was enabled"
            layout_graphic = semantic_graphic.layout_graphic
            assert layout_graphic
            assert layout_graphic.local_file_path
            zip_file.write(
                layout_graphic.local_file_path,
                semantic_graphic.relative_path
            )
    LOGGER.debug('response_content (bytes): %d', Path(zip_filename).stat().st_size)


def get_xml_tree(xml_root: etree.ElementBase) -> etree._ElementTree:
    if isinstance(xml_root, etree._ElementTree):  # pylint: disable=protected-access
        # Note: _XSLTResultTree is extending _ElementTree
        return xml_root
    return etree.ElementTree(xml_root)


def serialize_xml_to_file(
    xml_root: etree.ElementBase,
    filename: str
):
    get_xml_tree(xml_root).write(
        filename,
        encoding='utf-8',
        pretty_print=False
    )


@dataclass
class DocumentRequestParameters:
    first_page: Optional[int] = None
    last_page: Optional[int] = None


class ScienceBeamParserError(RuntimeError):
    pass


class BadRequestScienceBeamParserError(ScienceBeamParserError):
    pass


class UnsupportedRequestMediaTypeScienceBeamParserError(BadRequestScienceBeamParserError):
    pass


class UnsupportedResponseMediaTypeScienceBeamParserError(BadRequestScienceBeamParserError):
    pass


class ScienceBeamBaseParser:
    def __init__(self, config: AppConfig):
        self.config = config
        self.download_manager = DownloadManager(
            download_dir=get_download_dir(config)
        )
        self.pdfalto_wrapper = PdfAltoWrapper(
            self.download_manager.download_if_url(config['pdfalto']['path'])
        )
        self.pdfalto_wrapper.ensure_executable()
        self.app_context = AppContext(
            app_config=config,
            download_manager=self.download_manager,
            lazy_wapiti_binary_wrapper=LazyWapitiBinaryWrapper(
                install_url=config.get('wapiti', {}).get('install_source'),
                download_manager=self.download_manager
            )
        )
        self.fulltext_processor_config = FullTextProcessorConfig.from_app_config(app_config=config)
        self.fulltext_models = load_models(
            config,
            app_context=self.app_context,
            fulltext_processor_config=self.fulltext_processor_config
        )
        if config.get('preload_on_startup'):
            self.fulltext_models.preload()
        self.app_features_context = load_app_features_context(
            config,
            download_manager=self.download_manager
        )
        tei_to_jats_config = config.get('xslt', {}).get('tei_to_jats', {})
        self.tei_to_jats_xslt_transformer = XsltTransformerWrapper.from_template_file(
            TEI_TO_JATS_XSLT_FILE,
            xslt_template_parameters=tei_to_jats_config.get('parameters', {})
        )
        self.doc_to_pdf_enabled = config.get('doc_to_pdf', {}).get('enabled', True)
        self.doc_to_pdf_convert_parameters = config.get('doc_to_pdf', {}).get('convert', {})
        self.doc_converter_wrapper = DocConverterWrapper(
            **config.get('doc_to_pdf', {}).get('listener', {})
        )


class ScienceBeamParserBaseSession:
    def __init__(
        self,
        parser: 'ScienceBeamParser',
        temp_dir: Optional[str] = None,
        fulltext_processor_config: Optional[FullTextProcessorConfig] = None,
        document_request_parameters: Optional[DocumentRequestParameters] = None
    ):
        self.parser = parser
        self.exit_stack = ExitStack()
        self._temp_dir: Optional[str] = temp_dir
        if fulltext_processor_config is None:
            fulltext_processor_config = parser.fulltext_processor_config
        self.fulltext_processor_config = fulltext_processor_config
        if document_request_parameters is None:
            document_request_parameters = DocumentRequestParameters()
        self.document_request_parameters = document_request_parameters

    def __enter__(self) -> 'ScienceBeamParserBaseSession':
        return self

    def close(self):
        self.exit_stack.close()

    def __exit__(self, exc, value, tb):
        self.close()

    @property
    def temp_dir(self) -> str:
        if not self._temp_dir:
            temp_dir_context = TemporaryDirectory(  # pylint: disable=consider-using-with
                suffix='-sb-parser'
            )
            self.exit_stack.push(temp_dir_context)
            self._temp_dir = temp_dir_context.__enter__()
        return self._temp_dir

    @property
    def temp_path(self) -> Path:
        return Path(self.temp_dir)


class _ScienceBeamParserSessionDerivative:
    def __init__(
        self,
        session: 'ScienceBeamParserBaseSession'
    ):
        self.session = session

    @property
    def parser(self) -> ScienceBeamBaseParser:
        return self.session.parser

    @property
    def temp_dir(self) -> str:
        return self.session.temp_dir

    @property
    def temp_path(self) -> Path:
        return self.session.temp_path


class ScienceBeamParserSessionParsedSemanticDocument(_ScienceBeamParserSessionDerivative):
    def __init__(
        self,
        session: 'ScienceBeamParserBaseSession',
        semantic_document: SemanticDocument
    ):
        super().__init__(session=session)
        self.semantic_document = semantic_document

    @property
    def tei_to_jats_xslt_transformer(self) -> XsltTransformerWrapper:
        return self.parser.tei_to_jats_xslt_transformer

    def _get_tei_to_jats_xml_root(self, xml_root: etree.ElementBase) -> etree.ElementBase:
        start = monotonic()
        xml_root = self.tei_to_jats_xslt_transformer(xml_root)
        end = monotonic()
        LOGGER.info('tei to jats, took=%.3fs', end - start)
        return xml_root

    def _serialize_xml_to_file(
        self,
        xml_root: etree.ElementBase,
        filename: str
    ) -> str:
        start = monotonic()
        serialize_xml_to_file(xml_root, filename=filename)
        end = monotonic()
        LOGGER.info('serializing xml, took=%.3fs', end - start)
        return filename

    def get_supported_response_media_type(self) -> Set[str]:
        return {
            MediaTypes.TEI_XML,
            MediaTypes.TEI_ZIP,
            MediaTypes.JATS_XML,
            MediaTypes.JATS_ZIP
        }

    def get_local_file_for_response_media_type(
        self,
        response_media_type: str
    ) -> str:
        if response_media_type not in self.get_supported_response_media_type():
            raise UnsupportedResponseMediaTypeScienceBeamParserError()
        tei_document = get_tei_for_semantic_document(
            self.semantic_document
        )
        xml_root = tei_document.root
        relative_xml_filename = 'tei.xml'
        if response_media_type in JATS_MEDIA_TYPES:
            xml_root = self._get_tei_to_jats_xml_root(xml_root)
            relative_xml_filename = 'jats.xml'
        local_xml_filename = os.path.join(self.temp_dir, relative_xml_filename)
        self._serialize_xml_to_file(xml_root, local_xml_filename)
        LOGGER.debug('local_xml_filename: %r', local_xml_filename)
        if response_media_type in ASSET_ZIP_MEDIA_TYPES:
            zip_filename = os.path.join(self.temp_dir, 'results.zip')
            create_asset_zip_for_semantic_document(
                zip_filename,
                semantic_document=self.semantic_document,
                local_xml_filename=local_xml_filename,
                relative_xml_filename=relative_xml_filename
            )
            return zip_filename
        return local_xml_filename


class ScienceBeamParserSessionParsedLayoutDocument(_ScienceBeamParserSessionDerivative):
    def __init__(
        self,
        session: 'ScienceBeamParserBaseSession',
        layout_document: LayoutDocument,
        pdf_path: str
    ):
        super().__init__(session=session)
        self.layout_document = layout_document
        self.pdf_path = pdf_path

    @property
    def fulltext_models(self) -> FullTextModels:
        return self.parser.fulltext_models

    @property
    def app_features_context(self) -> AppFeaturesContext:
        return self.parser.app_features_context

    def _get_semantic_document(
        self,
        fulltext_processor: FullTextProcessor
    ) -> SemanticDocument:
        context = FullTextProcessorDocumentContext(
            pdf_path=self.pdf_path,
            temp_dir=self.temp_dir
        )
        semantic_document = (
            fulltext_processor
            .get_semantic_document_for_layout_document(
                self.layout_document,
                context=context
            )
        )
        return semantic_document

    def get_parsed_semantic_document(
        self,
        fulltext_processor_config: Optional[FullTextProcessorConfig] = None
    ) -> ScienceBeamParserSessionParsedSemanticDocument:
        if fulltext_processor_config is None:
            fulltext_processor_config = self.session.fulltext_processor_config
        fulltext_processor = FullTextProcessor(
            self.fulltext_models,
            app_features_context=self.app_features_context,
            config=fulltext_processor_config
        )
        return ScienceBeamParserSessionParsedSemanticDocument(
            self.session,
            self._get_semantic_document(fulltext_processor)
        )

    def get_local_file_for_response_media_type(
        self,
        response_media_type: str
    ) -> str:
        if response_media_type == MediaTypes.PDF:
            return self.pdf_path
        fulltext_processor_config = self.session.fulltext_processor_config
        if response_media_type in ASSET_ZIP_MEDIA_TYPES:
            fulltext_processor_config = (
                fulltext_processor_config
                ._replace(
                    extract_graphic_assets=True,
                    extract_graphic_bounding_boxes=True
                )
            )
            assert fulltext_processor_config.extract_graphic_assets, \
                "extract_graphic_assets required for asset zip"
        return (
            self.get_parsed_semantic_document(
                fulltext_processor_config
            ).get_local_file_for_response_media_type(
                response_media_type
            )
        )


class ScienceBeamParserSessionSource(_ScienceBeamParserSessionDerivative):
    def __init__(
        self,
        session: 'ScienceBeamParserBaseSession',
        source_path: str,
        source_media_type: str
    ):
        super().__init__(session=session)
        self.source_path = source_path
        self.source_media_type = source_media_type
        self.lazy_pdf_path = LazyLoaded[str](self._get_or_convert_to_pdf_path)
        self.lazy_alto_xml_path = LazyLoaded[str](self._parse_to_alto_xml)
        self.lazy_parsed_layout_document = LazyLoaded[
            ScienceBeamParserSessionParsedLayoutDocument
        ](self._parse_to_parsed_layout_document)

    @property
    def parser(self) -> ScienceBeamBaseParser:
        return self.session.parser

    @property
    def doc_to_pdf_enabled(self) -> bool:
        return self.parser.doc_to_pdf_enabled

    @property
    def doc_converter_wrapper(self) -> DocConverterWrapper:
        return self.parser.doc_converter_wrapper

    @property
    def doc_to_pdf_convert_parameters(self) -> dict:
        return self.parser.doc_to_pdf_convert_parameters

    @property
    def pdfalto_wrapper(self) -> PdfAltoWrapper:
        return self.parser.pdfalto_wrapper

    @property
    def document_request_parameters(self) -> DocumentRequestParameters:
        return self.session.document_request_parameters

    def _get_or_convert_to_pdf_path(
        self
    ) -> str:
        LOGGER.info(
            'media_type=%r (filename=%r)',
            self.source_media_type,
            self.source_path
        )
        if self.source_media_type in DOC_TO_PDF_SUPPORTED_MEDIA_TYPES:
            if not self.doc_to_pdf_enabled:
                LOGGER.info('doc to pdf not enabled')
                raise UnsupportedRequestMediaTypeScienceBeamParserError(
                    'doc to pdf not enabled'
                )
            target_temp_file = self.doc_converter_wrapper.convert(
                self.source_path,
                **self.doc_to_pdf_convert_parameters
            )
            return target_temp_file
        if self.source_media_type != MediaTypes.PDF:
            raise UnsupportedRequestMediaTypeScienceBeamParserError(
                'unsupported media type: %r' % self.source_media_type
            )
        return self.source_path

    def _parse_to_alto_xml(self) -> str:
        output_path = os.path.join(self.temp_dir, TEMP_ALTO_XML_FILENAME)
        self.pdfalto_wrapper.convert_pdf_to_pdfalto_xml(
            str(self.lazy_pdf_path.get()),
            str(output_path),
            first_page=self.document_request_parameters.first_page,
            last_page=self.document_request_parameters.last_page
        )
        return output_path

    def _parse_to_parsed_layout_document(
        self
    ) -> ScienceBeamParserSessionParsedLayoutDocument:
        pdf_path = self.lazy_pdf_path.get()
        root = etree.parse(self.lazy_alto_xml_path.get())
        layout_document = normalize_layout_document(
            parse_alto_root(root),
            preserve_empty_pages=True
        )
        return ScienceBeamParserSessionParsedLayoutDocument(
            self.session,
            layout_document=layout_document,
            pdf_path=pdf_path
        )

    def get_parsed_layout_document(self) -> ScienceBeamParserSessionParsedLayoutDocument:
        return self.lazy_parsed_layout_document.get()

    def get_layout_document(self) -> LayoutDocument:
        return self.get_parsed_layout_document().layout_document

    def get_local_file_for_response_media_type(
        self,
        response_media_type: str
    ) -> str:
        if response_media_type == MediaTypes.PDF:
            return self.lazy_pdf_path.get()
        if response_media_type == MediaTypes.ALTO_XML:
            return self.lazy_alto_xml_path.get()
        return self.get_parsed_layout_document().get_local_file_for_response_media_type(
            response_media_type
        )


class ScienceBeamParserSession(ScienceBeamParserBaseSession):
    def __enter__(self) -> 'ScienceBeamParserSession':
        super().__enter__()
        return self

    def get_source(
        self,
        source_path: str,
        source_media_type: str
    ) -> ScienceBeamParserSessionSource:
        return ScienceBeamParserSessionSource(
            self,
            source_path=source_path,
            source_media_type=source_media_type
        )


class ScienceBeamParser(ScienceBeamBaseParser):
    @staticmethod
    def from_config(config: AppConfig) -> 'ScienceBeamParser':
        return ScienceBeamParser(config)

    def get_new_session(self, **kwargs) -> ScienceBeamParserSession:
        return ScienceBeamParserSession(self, **kwargs)



import logging
from typing import Iterable, Optional, Tuple

from sciencebeam_parser.utils.misc import iter_ids
from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticHeading,
    SemanticLabel,
    SemanticNote,
    SemanticRawReference,
    SemanticRawReferenceText
)
from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.models.extract import ModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


def is_looks_like_reference(layout_block: LayoutBlock) -> bool:
    # a quick and dirty check whether this remotely looks like a reference
    return len(list(layout_block.iter_all_tokens())) > 3


class ReferenceSegmenterSemanticExtractor(ModelSemanticExtractor):
    def iter_semantic_content_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        ids_iterator = iter(iter_ids('b'))
        ref: Optional[SemanticRawReference] = None
        is_first_ref = True
        for name, layout_block in entity_tokens:
            if name == '<label>':
                if not ref:
                    ref = SemanticRawReference(content_id=next(ids_iterator, '?'))
                ref.add_content(SemanticLabel(layout_block=layout_block))
                continue
            if name == '<reference>':
                if not ref and is_first_ref and not is_looks_like_reference(layout_block):
                    yield SemanticHeading(layout_block=layout_block)
                    is_first_ref = False
                    continue
                if not ref:
                    ref = SemanticRawReference(content_id=next(ids_iterator, '?'))
                ref.add_content(SemanticRawReferenceText(layout_block=layout_block))
                yield ref
                ref = None
                is_first_ref = False
                continue
            yield SemanticNote(layout_block=layout_block, note_type=name)
        if ref:
            yield ref



import argparse
import logging
import os
from logging.config import dictConfig

from flask import Flask

from sciencebeam_parser.app.parser import ScienceBeamParser
from sciencebeam_parser.config.config import AppConfig
from sciencebeam_parser.service.blueprints.index import IndexBlueprint
from sciencebeam_parser.service.blueprints.api import ApiBlueprint
from sciencebeam_parser.resources.default_config import DEFAULT_CONFIG_FILE


LOGGER = logging.getLogger(__name__)


def create_app_for_parser(
    sciencebeam_parser: ScienceBeamParser
):
    app = Flask(__name__)

    index = IndexBlueprint()
    app.register_blueprint(index, url_prefix='/')

    api = ApiBlueprint(sciencebeam_parser)
    app.register_blueprint(api, url_prefix='/api')

    return app


def create_app_for_config(config: AppConfig):
    return create_app_for_parser(
        ScienceBeamParser.from_config(config)
    )


def create_app(config: AppConfig):
    return create_app_for_config(config)


def parse_args(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--host', required=False,
        help='Host to bind server to.'
    )
    parser.add_argument(
        '--port', type=int, default=8080,
        help='The port to listen to.'
    )
    parsed_args = parser.parse_args(argv)
    return parsed_args


def main(argv=None):
    args = parse_args(argv)
    config = AppConfig.load_yaml(DEFAULT_CONFIG_FILE).apply_environment_variables()
    logging_config = config.get('logging')
    if logging_config:
        for handler_config in logging_config.get('handlers', {}).values():
            filename = handler_config.get('filename')
            if not filename:
                continue
            dirname = os.path.dirname(filename)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
        try:
            dictConfig(logging_config)
        except ValueError:
            LOGGER.info('logging_config: %r', logging_config)
            raise
    LOGGER.info('app config: %s', config)
    app = create_app_for_config(config)
    app.run(port=args.port, host=args.host, threaded=True)


if __name__ == "__main__":
    logging.basicConfig(level='INFO')
    main()



import functools
from typing import cast

from sciencebeam_parser.app.context import AppContext
from sciencebeam_parser.models.model_impl import ModelImpl, T_ModelImplFactory


class EngineNames:
    DELFT = 'delft'
    WAPITI = 'wapiti'


def get_delft_model_impl_for_path(path: str, app_context: AppContext) -> ModelImpl:
    from sciencebeam_parser.models.delft_model_impl import (  # noqa pylint: disable=import-outside-toplevel
        DelftModelImpl
    )
    return DelftModelImpl(path, app_context=app_context)


def get_wapiti_model_impl_for_path(path: str, app_context: AppContext) -> ModelImpl:
    from sciencebeam_parser.models.wapiti_model_impl import (  # noqa pylint: disable=import-outside-toplevel
        WapitiModelImpl
    )
    return WapitiModelImpl(path, app_context=app_context)


def get_engine_name_for_config(config: dict) -> str:
    engine_name = config.get('engine')
    if engine_name:
        return engine_name
    return EngineNames.DELFT


def get_model_impl_for_config(config: dict, app_context: AppContext):
    path = config['path']
    engine_name = get_engine_name_for_config(config)
    if engine_name == EngineNames.WAPITI:
        return get_wapiti_model_impl_for_path(path, app_context=app_context)
    if engine_name == EngineNames.DELFT:
        return get_delft_model_impl_for_path(path, app_context=app_context)
    raise RuntimeError('invalid engine name: %r' % engine_name)


def get_model_impl_factory_for_config(
    config: dict,
    app_context: AppContext
) -> T_ModelImplFactory:
    return cast(
        T_ModelImplFactory,
        functools.partial(get_model_impl_for_config, config, app_context=app_context)
    )



from time import perf_counter


class StopWatch:
    def __init__(self):
        self.start = perf_counter()

    def get_elapsed_seconds(self, reset=False):
        end = perf_counter()
        elapsed = end - self.start
        if reset:
            self.start = end
        return elapsed


class StopWatchRecorder:
    def __init__(self):
        self.stop_watch = StopWatch()
        self.recorded_timings = []
        self.started = None

    def stop(self):
        self.start(None)

    def start(self, name: str):
        elapsed = self.stop_watch.get_elapsed_seconds(reset=True)
        if self.started:
            self.recorded_timings.append((self.started, elapsed))
        self.started = name

    def __str__(self):
        total = ('total', sum(elapsed for _, elapsed in self.recorded_timings))
        return ', '.join(
            '%s: %.6fs' % (name, elapsed)
            for name, elapsed in self.recorded_timings + [total]
        )



from typing import NamedTuple


class BoundingRange(NamedTuple):
    start: float
    length: float

    def validate(self) -> 'BoundingRange':
        if self.length < 0:
            raise ValueError(f'length must not be less than zero, was: {self.length}')
        return self

    def intersection(self, other: 'BoundingRange') -> 'BoundingRange':
        intersection_start = max(self.start, other.start)
        intersection_end = min(self.start + self.length, other.start + other.length)
        return BoundingRange(
            intersection_start,
            max(0, intersection_end - intersection_start)
        )


class BoundingBox(NamedTuple):
    x: float
    y: float
    width: float
    height: float

    @property
    def right(self) -> float:
        return self.x + self.width

    @property
    def bottom(self) -> float:
        return self.y + self.height

    @property
    def area(self) -> float:
        return self.width * self.height

    def scale_by(self, rx: float, ry: float) -> 'BoundingBox':
        return BoundingBox(self.x * rx, self.y * ry, self.width * rx, self.height * ry)

    @property
    def x_range(self):
        return BoundingRange(self.x, self.width).validate()

    @property
    def y_range(self):
        return BoundingRange(self.y, self.height).validate()

    def is_empty(self) -> bool:
        return self.width == 0 or self.height == 0

    def intersection(self, other: 'BoundingBox') -> 'BoundingBox':
        intersection_x_range = self.x_range.intersection(other.x_range)
        intersection_y_range = self.y_range.intersection(other.y_range)
        return BoundingBox(
            intersection_x_range.start,
            intersection_y_range.start,
            intersection_x_range.length,
            intersection_y_range.length
        )

    def __bool__(self) -> bool:
        return not self.is_empty()

    def __eq__(self, other):
        if other is None:
            return False
        return super().__eq__(other)



import logging
from typing import (
    List
)

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticRawEquation,
    SemanticRawEquationContent
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    T_ElementChildrenList,
    TeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class RawEquationTeiElementFactory(TeiElementFactory):
    def get_tei_children_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> List[etree.ElementBase]:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticRawEquation)
        semantic_raw_equation = semantic_content
        children: T_ElementChildrenList = [
            context.get_default_attributes_for_semantic_content(semantic_raw_equation)
        ]
        pending_whitespace = ''
        for child_semantic_content in semantic_raw_equation:
            if isinstance(child_semantic_content, SemanticRawEquationContent):
                layout_block = child_semantic_content.merged_block
                if pending_whitespace:
                    children.append(pending_whitespace)
                children.extend(context.iter_layout_block_tei_children(layout_block))
                pending_whitespace = layout_block.whitespace
                continue
            pending_whitespace = context.append_tei_children_list_and_get_whitespace(
                children,
                child_semantic_content,
                pending_whitespace=pending_whitespace
            )
        return [TEI_E('formula', *children)]



import logging
import re
from typing import Counter, Iterable, List, Optional, Set

from sciencebeam_parser.document.layout_document import (
    LayoutBlock,
    LayoutDocument,
    LayoutLine,
    LayoutToken
)
from sciencebeam_parser.models.data import (
    ContextAwareLayoutTokenFeatures,
    DocumentFeaturesContext,
    ModelDataGenerator,
    LayoutModelData,
    feature_linear_scaling_int,
    _LINESCALE,
    get_str_bool_feature_value
)


LOGGER = logging.getLogger(__name__)


NBSP = '\u00A0'


def format_feature_text(text: str) -> str:
    return re.sub(" |\t", NBSP, text.strip())


NBBINS_POSITION = 12

EMPTY_LAYOUT_TOKEN = LayoutToken('')
EMPTY_LAYOUT_LINE = LayoutLine([])


def get_block_status(line_index: int, line_count: int) -> str:
    return (
        'BLOCKSTART' if line_index == 0
        else (
            'BLOCKEND'
            if line_index == line_count - 1
            else 'BLOCKIN'
        )
    )


def get_page_status(
    block_index: int, block_count: int,
    is_first_block_token: bool,
    is_last_block_token: bool
) -> str:
    return (
        'PAGESTART' if block_index == 0 and is_first_block_token
        else (
            'PAGEEND'
            if block_index == block_count - 1 and is_last_block_token
            else 'PAGEIN'
        )
    )


# based on:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/features/FeatureFactory.java#L359-L367
def get_text_pattern(text: str) -> str:
    # Note: original code is meant to shadow numbers but are actually removed
    return re.sub(r'[^a-zA-Z ]', '', text).lower()


class SegmentationLineFeatures(ContextAwareLayoutTokenFeatures):
    def __init__(
        self,
        document_features_context: DocumentFeaturesContext,
        layout_token: LayoutToken = EMPTY_LAYOUT_TOKEN
    ):
        super().__init__(
            layout_token,
            document_features_context=document_features_context,
            layout_line=EMPTY_LAYOUT_LINE
        )
        self.line_text = ''
        self.second_token_text = ''
        self.page_blocks: List[LayoutBlock] = []
        self.page_block_index: int = 0
        self.block_lines: List[LayoutLine] = []
        self.block_line_index: int = 0
        self.previous_layout_token: Optional[LayoutToken] = None
        self.max_block_line_text_length = 0
        self.document_token_count = 0
        self.document_token_index = 0
        self.is_repetitive_pattern: bool = False
        self.is_first_repetitive_pattern: bool = False

    def get_block_status(self) -> str:
        return get_block_status(self.block_line_index, len(self.block_lines))

    def get_page_status(self) -> str:
        return get_page_status(
            self.page_block_index, len(self.page_blocks),
            is_first_block_token=self.block_line_index == 0,
            is_last_block_token=self.block_line_index == len(self.block_lines) - 1
        )

    def get_formatted_whole_line_feature(self) -> str:
        return format_feature_text(self.line_text)

    def get_str_is_repetitive_pattern(self) -> str:
        return get_str_bool_feature_value(self.is_repetitive_pattern)

    def get_str_is_first_repetitive_pattern(self) -> str:
        return get_str_bool_feature_value(self.is_first_repetitive_pattern)

    def get_dummy_str_is_repetitive_pattern(self) -> str:
        return '0'

    def get_dummy_str_is_first_repetitive_pattern(self) -> str:
        return '0'

    def get_dummy_str_is_main_area(self) -> str:
        # whether the block's bounding box intersects with the page bounding box
        return '1'

    def get_str_block_relative_line_length_feature(self) -> str:
        return str(feature_linear_scaling_int(
            len(self.line_text),
            self.max_block_line_text_length,
            _LINESCALE
        ))

    def get_str_relative_document_position(self) -> str:
        return str(feature_linear_scaling_int(
            self.document_token_index,
            self.document_token_count,
            NBBINS_POSITION
        ))


class SegmentationLineFeaturesProvider:
    def __init__(
        self,
        document_features_context: DocumentFeaturesContext,
        use_first_token_of_block: bool
    ):
        self.document_features_context = document_features_context
        self.use_first_token_of_block = use_first_token_of_block

    def iter_line_features(  # pylint: disable=too-many-locals
        self,
        layout_document: LayoutDocument
    ) -> Iterable[SegmentationLineFeatures]:
        segmentation_line_features = SegmentationLineFeatures(
            document_features_context=self.document_features_context
        )
        previous_token: Optional[LayoutToken] = None
        segmentation_line_features.document_token_count = sum(
            len(line.tokens)
            for block in layout_document.iter_all_blocks()
            for line in block.lines
        )
        pattern_candididate_block_iterable = (
            block
            for page in layout_document.pages
            for block_index, block in enumerate(page.blocks)
            if block_index < 2 or block_index > len(page.blocks) - 2
        )
        pattern_candididate_line_iterable = (
            block.lines[0]
            for block in pattern_candididate_block_iterable
            if block.lines and block.lines[0].tokens
        )
        all_pattern_by_line_id = {
            id(line): get_text_pattern(line.text)
            for line in pattern_candididate_line_iterable
        }
        LOGGER.debug('all_pattern_by_line_id: %s', all_pattern_by_line_id)
        pattern_by_line_id = {
            key: value
            for key, value in all_pattern_by_line_id.items()
            if len(value) >= 8  # Java GROBID sometimes counts an additional trailing space
        }
        pattern_counter = Counter(pattern_by_line_id.values())
        LOGGER.debug('pattern_counter: %s', pattern_counter)
        seen_repetitive_patterns: Set[str] = set()
        document_token_index = 0
        for page in layout_document.pages:
            blocks = page.blocks
            segmentation_line_features.page_blocks = blocks
            for block_index, block in enumerate(blocks):
                segmentation_line_features.page_block_index = block_index
                block_lines = block.lines
                segmentation_line_features.block_lines = block_lines
                block_line_texts = [line.text for line in block_lines]
                max_block_line_text_length = max(len(text) for text in block_line_texts)
                first_block_token = next(iter(block.iter_all_tokens()), None)
                assert first_block_token
                for line_index, line in enumerate(block_lines):
                    segmentation_line_features.document_token_index = document_token_index
                    document_token_index += len(line.tokens)
                    segmentation_line_features.layout_line = line
                    segmentation_line_features.block_line_index = line_index
                    segmentation_line_features.max_block_line_text_length = (
                        max_block_line_text_length
                    )
                    line_text = block_line_texts[line_index]
                    retokenized_token_texts = re.split(r" |\t|\f|\u00A0", line_text)
                    if not retokenized_token_texts:
                        continue
                    if self.use_first_token_of_block:
                        # Java GROBID uses the first token in the block
                        token = first_block_token
                    else:
                        token = line.tokens[0]
                    segmentation_line_features.layout_token = token
                    segmentation_line_features.line_text = line_text
                    segmentation_line_features.concatenated_line_tokens_text = line_text
                    segmentation_line_features.token_text = retokenized_token_texts[0].strip()
                    segmentation_line_features.second_token_text = (
                        retokenized_token_texts[1] if len(retokenized_token_texts) >= 2 else ''
                    )
                    segmentation_line_features.previous_layout_token = previous_token
                    line_pattern = pattern_by_line_id.get(id(line), '')
                    LOGGER.debug('line_pattern: %r', line_pattern)
                    segmentation_line_features.is_repetitive_pattern = (
                        pattern_counter[line_pattern] > 1
                    )
                    segmentation_line_features.is_first_repetitive_pattern = (
                        segmentation_line_features.is_repetitive_pattern
                        and line_pattern not in seen_repetitive_patterns
                    )
                    if segmentation_line_features.is_first_repetitive_pattern:
                        seen_repetitive_patterns.add(line_pattern)
                    yield segmentation_line_features
                    previous_token = token


class SegmentationDataGenerator(ModelDataGenerator):
    def __init__(
        self,
        document_features_context: DocumentFeaturesContext,
        use_first_token_of_block: bool
    ):
        self.document_features_context = document_features_context
        self.use_first_token_of_block = use_first_token_of_block

    def iter_model_data_for_layout_document(
        self,
        layout_document: LayoutDocument
    ) -> Iterable[LayoutModelData]:
        features_provider = SegmentationLineFeaturesProvider(
            document_features_context=self.document_features_context,
            use_first_token_of_block=self.use_first_token_of_block
        )
        for features in features_provider.iter_line_features(
            layout_document
        ):
            line_features: List[str] = [
                features.token_text,
                features.second_token_text or features.token_text,
                features.get_lower_token_text(),
                features.get_prefix(1),
                features.get_prefix(2),
                features.get_prefix(3),
                features.get_prefix(4),
                features.get_block_status(),
                features.get_page_status(),
                features.get_token_font_status(),
                features.get_token_font_size_feature(),
                features.get_str_is_bold(),
                features.get_str_is_italic(),
                features.get_capitalisation_status_using_allcap(),
                features.get_digit_status_using_containsdigits(),
                features.get_str_is_single_char(),
                features.get_dummy_str_is_proper_name(),
                features.get_dummy_str_is_common_name(),
                features.get_dummy_str_is_first_name(),
                features.get_dummy_str_is_year(),
                features.get_dummy_str_is_month(),
                features.get_dummy_str_is_email(),
                features.get_dummy_str_is_http(),
                features.get_str_relative_document_position(),
                features.get_dummy_str_relative_page_position(),
                features.get_line_punctuation_profile(),
                features.get_line_punctuation_profile_length_feature(),
                features.get_str_block_relative_line_length_feature(),
                features.get_dummy_str_is_bitmap_around(),
                features.get_dummy_str_is_vector_around(),
                features.get_str_is_repetitive_pattern(),
                features.get_str_is_first_repetitive_pattern(),
                features.get_dummy_str_is_main_area(),
                features.get_formatted_whole_line_feature()
            ]

            if len(line_features) != 34:
                raise AssertionError(
                    'expected features to have 34 features, but was=%d (features=%s)' % (
                        len(line_features), line_features
                    )
                )
            yield LayoutModelData(
                layout_line=features.layout_line,
                data_line=' '.join(line_features)
            )



import logging
from time import monotonic
from typing import Iterable, NamedTuple, Optional, Sequence

import pdf2image


LOGGER = logging.getLogger(__name__)


DEFAULT_PDF_RENDER_DPI = 200


class DocumentPageImage(NamedTuple):
    page_number: int
    page_image_path: str


def iter_pdf_document_page_images(
    pdf_path: str,
    output_dir: str,
    thread_count: int = 1,
    dpi: float = DEFAULT_PDF_RENDER_DPI,
    page_numbers: Optional[Sequence[int]] = None
) -> Iterable[DocumentPageImage]:
    first_page: int = 1
    last_page: Optional[int] = None
    if page_numbers:
        first_page = min(page_numbers)
        last_page = max(page_numbers)
    render_start = monotonic()
    output_files = pdf2image.convert_from_path(
        pdf_path=pdf_path,
        first_page=first_page,
        last_page=last_page,
        dpi=dpi,
        paths_only=True,
        output_folder=output_dir,
        thread_count=thread_count
    )
    render_end = monotonic()
    LOGGER.info(
        'rendered PDF pages, took=%.3fs, page_count=%d, dpi=%s, thread_count=%d',
        render_end - render_start,
        len(output_files),
        dpi,
        thread_count
    )
    for page_index, output_file in enumerate(output_files):
        page_number = first_page + page_index
        if page_numbers is not None and page_number not in page_numbers:
            LOGGER.debug('ignoring deselected page: %d (%s)', page_number, page_numbers)
            continue
        yield DocumentPageImage(
            page_number=page_number,
            page_image_path=output_file
        )



import logging
import re
from typing import Iterable, Mapping, Optional, Tuple

from sciencebeam_parser.document.semantic_document import (
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticRawAddress,
    SemanticRawAffiliation,
    SemanticRawAffiliationAddress,
    SemanticTitle,
    SemanticAbstract,
    SemanticRawAuthors
)
from sciencebeam_parser.document.layout_document import LayoutBlock, LayoutTokensText
from sciencebeam_parser.models.extract import SimpleModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


# based on:
#   grobid-core/src/main/java/org/grobid/core/data/BiblioItem.java
ABSTRACT_REGEX = r'^(?:(?:abstract|summary|résumé|abrégé|a b s t r a c t)(?:[.:])?)?\s*(.*)'


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<author>': SemanticRawAuthors,
    '<affiliation>': SemanticRawAffiliation,
    '<address>': SemanticRawAddress
}


def get_cleaned_abstract_text(text: str) -> str:
    if not text:
        return text
    m = re.match(ABSTRACT_REGEX, text, re.IGNORECASE)
    if not m:
        LOGGER.debug('text does not match regex: %r', text)
        return text
    return m.group(1)


def get_cleaned_abstract_layout_block(layout_block: LayoutBlock) -> LayoutBlock:
    if not layout_block or not layout_block.lines:
        return layout_block
    layout_tokens_text = LayoutTokensText(layout_block)
    text = str(layout_tokens_text)
    m = re.match(ABSTRACT_REGEX, text, re.IGNORECASE)
    if not m:
        LOGGER.debug('text does not match regex: %r', text)
        return layout_block
    start = m.start(1)
    LOGGER.debug('start: %d (text: %r)', start, text)
    return LayoutBlock.for_tokens(list(
        layout_tokens_text.iter_layout_tokens_between(start, len(text))
    ))


class HeaderSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def iter_semantic_content_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        has_title: bool = False
        has_abstract: bool = False
        aff_address: Optional[SemanticRawAffiliationAddress] = None
        next_previous_label: str = ''
        for name, layout_block in entity_tokens:
            previous_label = next_previous_label
            next_previous_label = name
            if name == '<title>' and not has_title:
                yield SemanticTitle(layout_block=layout_block)
                has_title = True
                continue
            if name == '<abstract>' and not has_abstract:
                abstract_layout_block = get_cleaned_abstract_layout_block(
                    layout_block
                )
                yield SemanticAbstract(layout_block=abstract_layout_block)
                has_abstract = True
                continue
            if name in {'<affiliation>', '<address>'}:
                if (
                    aff_address is not None
                    and name == '<affiliation>'
                    and previous_label in {'<affiliation>', '<address>'}
                ):
                    yield aff_address
                    aff_address = None
                if aff_address is None:
                    aff_address = SemanticRawAffiliationAddress()
                aff_address.add_content(self.get_semantic_content_for_entity_name(
                    name, layout_block
                ))
                continue
            if aff_address is not None:
                yield aff_address
                aff_address = None
            yield self.get_semantic_content_for_entity_name(
                name, layout_block
            )
        if aff_address is not None:
            yield aff_address



import logging

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.name.data import NameDataGenerator
from sciencebeam_parser.models.name.extract import NameSemanticExtractor
from sciencebeam_parser.models.name.training_data import (
    NameTeiTrainingDataGenerator,
    NameTrainingTeiParser
)
from sciencebeam_parser.models.model import Model


LOGGER = logging.getLogger(__name__)


class NameModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> NameDataGenerator:
        return NameDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> NameSemanticExtractor:
        return NameSemanticExtractor()

    def get_tei_training_data_generator(self) -> NameTeiTrainingDataGenerator:
        return NameTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> NameTrainingTeiParser:
        return NameTrainingTeiParser()



import logging
import re
from itertools import zip_longest
from typing import Mapping, NamedTuple, Optional, Sequence, Tuple, Union

from lxml import etree
from lxml.builder import ElementMaker


LOGGER = logging.getLogger(__name__)


class TagExpression(NamedTuple):
    tag: str
    attrib: Mapping[str, str]

    def create_node(self, *args, element_maker: ElementMaker):
        try:
            return element_maker(self.tag, self.attrib, *args)
        except ValueError as exc:
            raise ValueError(
                'failed to create node with tag=%r, attrib=%r due to %s' % (
                    self.tag, self.attrib, exc
                )
            ) from exc


def parse_tag_expression(tag_expression: str) -> TagExpression:
    match = re.match(r'^([^\[]+)(\[@?([^=]+)="(.+)"\])?$', tag_expression)
    if not match:
        raise ValueError('invalid tag expression: %s' % tag_expression)
    LOGGER.debug('match: %s', match.groups())
    tag_name = match.group(1)
    if match.group(2):
        attrib = {match.group(3): match.group(4)}
    else:
        attrib = {}
    return TagExpression(tag=tag_name, attrib=attrib)


def _get_last_child_or_none(element: etree.ElementBase) -> Optional[etree.ElementBase]:
    try:
        return element[-1]
    except IndexError:
        return None


def _append_text(element: etree.ElementBase, text: Optional[str]) -> None:
    if not text:
        return
    last_child = _get_last_child_or_none(element)
    if last_child is not None and last_child.tail:
        last_child.tail = last_child.tail + '' + text
    elif last_child is not None:
        last_child.tail = text
    elif element.text:
        element.text = element.text + '' + text
    else:
        element.text = text


def _get_common_path(path1: Sequence[str], path2: Sequence[str]) -> Sequence[str]:
    if path1 == path2:
        return path1
    common_path = []
    for path1_element, path2_element in zip_longest(path1, path2):
        if path1_element != path2_element:
            break
        common_path.append(path1_element)
    return common_path


def _get_element_at_path(
    current_element: etree.ElementBase,
    current_path: Sequence[str],
    required_path: Sequence[str],
    element_maker: ElementMaker
) -> Tuple[etree.ElementBase, Sequence[str]]:
    if required_path != current_path:
        common_path = _get_common_path(current_path, required_path)
        LOGGER.debug(
            'required element path: %s -> %s (common path: %s)',
            current_path, required_path, common_path
        )
        for _ in range(len(current_path) - len(common_path)):
            current_element = current_element.getparent()
        current_path = list(common_path)
        for path_fragment in required_path[len(common_path):]:
            try:
                parsed_path_fragment = parse_tag_expression(path_fragment)
                child = parsed_path_fragment.create_node(
                    element_maker=element_maker
                )
            except ValueError as exc:
                raise ValueError('failed to create node for %r due to %s' % (
                    path_fragment, exc
                )) from exc
            current_element.append(child)
            current_element = child
            current_path.append(path_fragment)
    return current_element, current_path


class XmlTreeWriter:
    def __init__(
        self,
        parent: etree.ElementBase,
        element_maker: ElementMaker
    ):
        self.current_element = parent
        self.current_path: Sequence[str] = []
        self.element_maker = element_maker

    @property
    def root(self) -> etree.ElementBase:
        return self.current_element.getroottree().getroot()

    def append_text(self, text: str):
        _append_text(self.current_element, text)

    def append(self, element_or_text: Union[etree.ElementBase, str]):
        if isinstance(element_or_text, str):
            self.append_text(element_or_text)
        else:
            self.current_element.append(element_or_text)

    def append_all(self, *element_or_text_list: Sequence[Union[etree.ElementBase, str]]):
        for element_or_text in element_or_text_list:
            self.append(element_or_text)

    def require_path(self, required_path: Sequence[str]):
        self.current_element, self.current_path = _get_element_at_path(
            self.current_element, self.current_path,
            required_path,
            element_maker=self.element_maker
        )

    def require_path_or_below(self, required_path: Sequence[str]):
        self.require_path(
            _get_common_path(self.current_path, required_path)
        )



import logging
import multiprocessing
from typing import (
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union
)

from sciencebeam_parser.models.data import AppFeaturesContext, DEFAULT_APP_FEATURES_CONTEXT
from sciencebeam_parser.models.model import LayoutDocumentLabelResult, Model
from sciencebeam_parser.cv_models.cv_model import ComputerVisionModel
from sciencebeam_parser.processors.fulltext.models import FullTextModels
from sciencebeam_parser.utils.misc import iter_ids

from sciencebeam_parser.document.semantic_document import (
    SemanticAffiliationAddress,
    SemanticAuthor,
    SemanticCitation,
    SemanticContentWrapper,
    SemanticDocument,
    SemanticEditor,
    SemanticFigure,
    SemanticFigureCitation,
    SemanticGraphic,
    SemanticInvalidReference,
    SemanticLabel,
    SemanticMixedContentWrapper,
    SemanticMixedNote,
    SemanticRawAffiliationAddress,
    SemanticRawAuthors,
    SemanticRawEditors,
    SemanticRawFigure,
    SemanticRawReference,
    SemanticRawReferenceText,
    SemanticRawTable,
    SemanticReference,
    SemanticReferenceCitation,
    SemanticReferenceList,
    SemanticSection,
    SemanticSectionTypes,
    SemanticTable,
    SemanticTableCitation,
    T_SemanticContentWrapper,
    T_SemanticName,
    T_SemanticRawNameList
)
from sciencebeam_parser.document.tei_document import TeiDocument, get_tei_for_semantic_document
from sciencebeam_parser.document.layout_document import LayoutDocument
from sciencebeam_parser.models.segmentation.model import SegmentationModel
from sciencebeam_parser.models.header.model import HeaderModel
from sciencebeam_parser.models.name.model import NameModel
from sciencebeam_parser.models.affiliation_address.model import AffiliationAddressModel
from sciencebeam_parser.models.fulltext.model import FullTextModel
from sciencebeam_parser.models.figure.model import FigureModel
from sciencebeam_parser.models.table.model import TableModel
from sciencebeam_parser.models.reference_segmenter.model import ReferenceSegmenterModel
from sciencebeam_parser.models.citation.model import CitationModel
from sciencebeam_parser.processors.ref_matching import (
    ChainedContentIdMatcher,
    ContentIdMatcher,
    PartialContentIdMatcher,
    SimpleContentIdMatcher
)
from sciencebeam_parser.processors.document_page_image import (
    iter_pdf_document_page_images
)
from sciencebeam_parser.processors.graphic_matching import (
    BoundingBoxDistanceGraphicMatcher,
    ChainedGraphicMatcher,
    GraphicMatcher,
    GraphicRelatedBlockTextGraphicMatcher,
    OpticalCharacterRecognitionGraphicMatcher
)
from sciencebeam_parser.processors.graphic_provider import (
    DocumentGraphicProvider,
    SimpleDocumentGraphicProvider,
    get_graphic_matching_candidate_page_numbers_for_semantic_content_list,
    get_layout_document_with_graphics_replaced_by_graphics,
    get_layout_document_with_text_and_graphics_replaced_by_graphics,
    get_page_numbers_with_mostly_bitmap_graphics,
    get_page_numbers_with_uncommon_page_dimension
)
from sciencebeam_parser.processors.fulltext.config import (
    FullTextProcessorConfig
)


LOGGER = logging.getLogger(__name__)


class FullTextProcessorDocumentContext(NamedTuple):
    pdf_path: Optional[str] = None
    temp_dir: Optional[str] = None


def get_cv_document_graphic_provider(
    cv_model: ComputerVisionModel,
    context: FullTextProcessorDocumentContext,
    page_numbers: Optional[Sequence[int]],
    cv_render_dpi: float
) -> DocumentGraphicProvider:
    from sciencebeam_parser.processors.cv_graphic_provider import (  # noqa pylint: disable=import-outside-toplevel
        ComputerVisionDocumentGraphicProvider
    )
    assert context.pdf_path
    assert context.temp_dir
    return ComputerVisionDocumentGraphicProvider(
        cv_model,
        iter_pdf_document_page_images(
            pdf_path=context.pdf_path,
            output_dir=context.temp_dir,
            page_numbers=page_numbers,
            dpi=cv_render_dpi,
            thread_count=multiprocessing.cpu_count()
        ),
        temp_dir=context.temp_dir
    )


class FullTextProcessor:
    def __init__(
        self,
        fulltext_models: FullTextModels,
        config: Optional[FullTextProcessorConfig] = None,
        app_features_context: AppFeaturesContext = DEFAULT_APP_FEATURES_CONTEXT
    ) -> None:
        self.fulltext_models = fulltext_models
        self.app_features_context = app_features_context
        if not config:
            config = FullTextProcessorConfig()
        self.config = config

    @property
    def segmentation_model(self) -> SegmentationModel:
        return self.fulltext_models.segmentation_model

    @property
    def header_model(self) -> HeaderModel:
        return self.fulltext_models.header_model

    @property
    def affiliation_address_model(self) -> AffiliationAddressModel:
        return self.fulltext_models.affiliation_address_model

    @property
    def name_header_model(self) -> NameModel:
        return self.fulltext_models.name_header_model

    @property
    def name_citation_model(self) -> NameModel:
        return self.fulltext_models.name_citation_model

    @property
    def fulltext_model(self) -> FullTextModel:
        return self.fulltext_models.fulltext_model

    @property
    def figure_model(self) -> FigureModel:
        return self.fulltext_models.figure_model

    @property
    def table_model(self) -> TableModel:
        return self.fulltext_models.table_model

    @property
    def reference_segmenter_model(self) -> ReferenceSegmenterModel:
        return self.fulltext_models.reference_segmenter_model

    @property
    def citation_model(self) -> CitationModel:
        return self.fulltext_models.citation_model

    def get_semantic_document_for_layout_document(
        self,
        layout_document: LayoutDocument,
        context: Optional[FullTextProcessorDocumentContext] = None
    ) -> SemanticDocument:
        if context is None:
            context = FullTextProcessorDocumentContext()
        layout_document = self._preprocess_layout_graphics(
            layout_document,
            context=context
        )
        segmentation_label_result = self.segmentation_model.get_label_layout_document_result(
            layout_document,
            app_features_context=self.app_features_context
        )
        header_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<header>'
        ).remove_empty_blocks()
        document = SemanticDocument()
        if self.config.extract_front:
            self._process_header_layout_document(
                header_layout_document=header_layout_document,
                semantic_document=document
            )

        if self.config.extract_body_sections:
            self._update_semantic_section_using_segmentation_result_and_fulltext_model(
                document.body_section,
                segmentation_label_result,
                '<body>',
                SemanticSectionTypes.OTHER
            )
        if self.config.extract_acknowledgements:
            self._update_semantic_section_using_segmentation_result_and_fulltext_model(
                document.back_section,
                segmentation_label_result,
                '<acknowledgement>',
                SemanticSectionTypes.ACKNOWLEDGEMENT
            )
        if self.config.extract_back_sections:
            self._update_semantic_section_using_segmentation_result_and_fulltext_model(
                document.back_section,
                segmentation_label_result,
                '<annex>',
                SemanticSectionTypes.OTHER
            )
        if self.config.extract_references:
            self._extract_raw_references_from_segmentation(
                semantic_document=document,
                segmentation_label_result=segmentation_label_result
            )
        if self.config.extract_citation_fields:
            self._extract_reference_fields_from_raw_references(
                semantic_document=document
            )
            if self.config.extract_citation_authors or self.config.extract_citation_editors:
                self._extract_reference_name_lists_from_raw_references(
                    semantic_document=document
                )
            references = list(document.iter_by_type_recursively(SemanticReference))
            ref_citations = list(document.iter_by_type_recursively(SemanticReferenceCitation))
            self._assign_content_ids(references, iter(iter_ids('b')))
            self._assign_target_content_ids(ref_citations, ChainedContentIdMatcher([
                SimpleContentIdMatcher(
                    self._get_semantic_content_text_by_content_id(references, SemanticLabel)
                ),
                PartialContentIdMatcher(
                    self._get_semantic_content_text_by_content_id(
                        references, SemanticRawReferenceText
                    )
                )
            ]))
        if self.config.extract_figure_fields:
            self._extract_figure_fields_from_raw_figures(semantic_document=document)
            figures = list(document.iter_by_type_recursively(SemanticFigure))
            figure_citations = list(document.iter_by_type_recursively(SemanticFigureCitation))
            self._assign_content_ids(figures, iter(iter_ids('fig_')))
            self._assign_target_content_ids(figure_citations, SimpleContentIdMatcher(
                self._get_semantic_content_text_by_content_id(figures, SemanticLabel)
            ))
        if self.config.extract_table_fields:
            self._extract_table_fields_from_raw_tables(semantic_document=document)
            tables = list(document.iter_by_type_recursively(SemanticTable))
            table_citations = list(document.iter_by_type_recursively(SemanticTableCitation))
            self._assign_content_ids(tables, iter(iter_ids('tab_')))
            self._assign_target_content_ids(table_citations, SimpleContentIdMatcher(
                self._get_semantic_content_text_by_content_id(tables, SemanticLabel)
            ))
        if self.config.extract_graphic_bounding_boxes:
            self._process_graphics(
                document=document,
                layout_document=layout_document,
                context=context
            )
        return document

    def _process_header_layout_document(
        self,
        header_layout_document: LayoutDocument,
        semantic_document: SemanticDocument
    ):
        LOGGER.debug('header_layout_document: %s', header_layout_document)
        if not header_layout_document.pages:
            return
        labeled_layout_tokens = self.header_model.predict_labels_for_layout_document(
            header_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        entity_blocks = self.header_model.iter_entity_layout_blocks_for_labeled_layout_tokens(
            labeled_layout_tokens
        )
        self.header_model.update_semantic_document_with_entity_blocks(
            semantic_document, entity_blocks
        )
        if self.config.extract_authors:
            self._process_raw_authors(semantic_document.front)
        if self.config.extract_affiliations:
            self._process_raw_affiliations(semantic_document)

    def _preprocess_layout_graphics(
        self,
        layout_document: LayoutDocument,
        context: FullTextProcessorDocumentContext
    ) -> LayoutDocument:
        if not self.config.use_cv_model:
            return layout_document
        candidate_page_numbers = sorted(
            set(get_page_numbers_with_uncommon_page_dimension(layout_document))
            - set(get_page_numbers_with_mostly_bitmap_graphics(layout_document))
        )
        LOGGER.debug('candidate_page_numbers: %r', candidate_page_numbers)
        if not candidate_page_numbers:
            return layout_document
        document_graphic_provider = self._get_document_graphic_provider(
            context=context,
            page_numbers=candidate_page_numbers
        )
        semantic_graphics = list(
            document_graphic_provider.iter_semantic_graphic_for_layout_document(
                layout_document,
                extract_graphic_assets=self.config.extract_graphic_assets
            )
        )
        if not semantic_graphics:
            LOGGER.info('no semantic graphics found on pages %r', candidate_page_numbers)
            return layout_document
        if not self.config.replace_text_by_cv_graphic:
            return get_layout_document_with_graphics_replaced_by_graphics(
                layout_document,
                semantic_graphics
            )
        return get_layout_document_with_text_and_graphics_replaced_by_graphics(
            layout_document,
            semantic_graphics
        )

    def _process_graphics(
        self,
        document: SemanticDocument,
        layout_document: LayoutDocument,
        context: FullTextProcessorDocumentContext
    ):
        unmatched_graphics_container = SemanticMixedNote(note_type='unmatched_graphics')
        candidate_semantic_content_list = list(
            document.iter_by_types_recursively((SemanticFigure, SemanticTable,))
        )
        self._match_graphic_elements(
            semantic_graphic_list=list(
                self._get_document_graphic_provider(
                    context=context,
                    page_numbers=(
                        get_graphic_matching_candidate_page_numbers_for_semantic_content_list(
                            candidate_semantic_content_list,
                            layout_document=layout_document
                        )
                    )
                ).iter_semantic_graphic_for_layout_document(
                    layout_document,
                    extract_graphic_assets=self.config.extract_graphic_assets
                )
            ),
            candidate_semantic_content_list=candidate_semantic_content_list,
            unmatched_graphics_container=unmatched_graphics_container
        )
        if not unmatched_graphics_container.is_empty():
            LOGGER.debug('unmatched_graphics_container: %r', unmatched_graphics_container)
            document.back_section.add_content(unmatched_graphics_container)
        else:
            LOGGER.debug('no unmatched graphics')

    def _get_document_graphic_provider(
        self,
        context: FullTextProcessorDocumentContext,
        page_numbers: Optional[Sequence[int]]
    ) -> DocumentGraphicProvider:
        if self.config.use_cv_model:
            assert self.fulltext_models.cv_model is not None
            return get_cv_document_graphic_provider(
                cv_model=self.fulltext_models.cv_model,
                context=context,
                page_numbers=page_numbers,
                cv_render_dpi=self.config.cv_render_dpi
            )
        return SimpleDocumentGraphicProvider()

    def _match_graphic_elements(
        self,
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper],
        unmatched_graphics_container: SemanticMixedContentWrapper
    ):
        _graphic_matchers: List[GraphicMatcher] = [
            BoundingBoxDistanceGraphicMatcher(
                max_distance=self.config.max_graphic_distance
            ),
            GraphicRelatedBlockTextGraphicMatcher()
        ]
        if self.config.use_ocr_model:
            assert self.fulltext_models.ocr_model
            _graphic_matchers.append(
                OpticalCharacterRecognitionGraphicMatcher(
                    ocr_model=self.fulltext_models.ocr_model
                )
            )
        graphic_matcher = ChainedGraphicMatcher(_graphic_matchers)
        graphic_match_result = graphic_matcher.get_graphic_matches(
            semantic_graphic_list=semantic_graphic_list,
            candidate_semantic_content_list=candidate_semantic_content_list
        )
        for graphic_match in graphic_match_result.graphic_matches:
            if isinstance(graphic_match.candidate_semantic_content,  SemanticMixedContentWrapper):
                graphic_match.candidate_semantic_content.add_content(
                    graphic_match.semantic_graphic
                )
        LOGGER.info('unmatched_graphics: %r', graphic_match_result.unmatched_graphics)
        for unmatched_graphic in graphic_match_result.unmatched_graphics:
            unmatched_graphics_container.add_content(unmatched_graphic)

    def _assign_content_ids(
        self,
        semantic_content_iterable: Iterable[SemanticMixedContentWrapper],
        content_id_iterator: Iterator[str]
    ):
        for semantic_content in semantic_content_iterable:
            semantic_content.content_id = next(content_id_iterator)

    def _get_semantic_content_text_by_content_id(
        self,
        semantic_content_iterable: Iterable[SemanticMixedContentWrapper],
        type_: Type[SemanticContentWrapper]
    ) -> Mapping[str, str]:
        d = {}
        for semantic_content in semantic_content_iterable:
            if not semantic_content.content_id:
                continue
            text = semantic_content.get_text_by_type(type_)
            if not text:
                continue
            d[semantic_content.content_id] = text
        return d

    def _assign_target_content_ids(
        self,
        semantic_content_iterable: Iterable[SemanticCitation],
        content_id_matcher: ContentIdMatcher
    ):
        for citation in semantic_content_iterable:
            content_id = content_id_matcher.get_id_by_text(citation.get_text())
            if content_id:
                citation.target_content_id = content_id

    def _process_raw_authors(self, semantic_parent: SemanticMixedContentWrapper):
        result_content: List[SemanticContentWrapper] = []
        raw_authors: List[SemanticRawAuthors] = []
        for semantic_content in semantic_parent:
            if isinstance(semantic_content, SemanticRawAuthors):
                raw_authors.append(semantic_content)
                continue
            result_content.append(semantic_content)
        if raw_authors:
            if self.config.merge_raw_authors:
                raw_authors_layout_documents = [
                    LayoutDocument.for_blocks([
                        block
                        for raw_author in raw_authors
                        for block in raw_author.iter_blocks()
                    ])
                ]
            else:
                raw_authors_layout_documents = [
                    LayoutDocument.for_blocks(list(raw_author.iter_blocks()))
                    for raw_author in raw_authors
                ]
            labeled_layout_tokens_list = self.name_header_model.predict_labels_for_layout_documents(
                raw_authors_layout_documents,
                app_features_context=self.app_features_context
            )
            LOGGER.debug('labeled_layout_tokens_list (author): %r', labeled_layout_tokens_list)
            authors_iterable = (
                author
                for labeled_layout_tokens in labeled_layout_tokens_list
                for author in (
                    self.name_header_model.iter_semantic_content_for_labeled_layout_tokens(
                        labeled_layout_tokens
                    )
                )
            )
            for author in authors_iterable:
                result_content.append(author)
        semantic_parent.mixed_content = result_content

    def _process_raw_affiliations(self, semantic_document: SemanticDocument):
        result_content: List[SemanticContentWrapper] = []
        raw_aff_address_list: List[SemanticRawAffiliationAddress] = []
        for semantic_content in semantic_document.front:
            if isinstance(semantic_content, SemanticRawAffiliationAddress):
                raw_aff_address_list.append(semantic_content)
                continue
            result_content.append(semantic_content)
        if raw_aff_address_list:
            raw_aff_layout_documents = [
                LayoutDocument.for_blocks(list(raw_aff_or_address.iter_blocks()))
                for raw_aff_or_address in raw_aff_address_list
            ]
            labeled_layout_tokens_list = (
                self.affiliation_address_model
                .predict_labels_for_layout_documents(
                    raw_aff_layout_documents,
                    app_features_context=self.app_features_context
                )
            )
            LOGGER.debug('labeled_layout_tokens_list (aff): %r', labeled_layout_tokens_list)
            aff_iterable = (
                aff
                for labeled_layout_tokens in labeled_layout_tokens_list
                for aff in (
                    self.affiliation_address_model
                    .iter_semantic_content_for_labeled_layout_tokens(labeled_layout_tokens)
                )
            )
            for aff in aff_iterable:
                result_content.append(aff)
        semantic_document.front.mixed_content = result_content
        self._assign_content_ids(
            semantic_document.front.iter_by_type(SemanticAffiliationAddress),
            iter(iter_ids('aff'))
        )

    def _extract_raw_references_from_segmentation(
        self,
        semantic_document: SemanticDocument,
        segmentation_label_result: LayoutDocumentLabelResult
    ):
        references_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<references>'
        ).remove_empty_blocks()
        LOGGER.debug('references_layout_document: %s', references_layout_document)
        if not references_layout_document:
            return
        labeled_layout_tokens = self.reference_segmenter_model.predict_labels_for_layout_document(
            references_layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens: %r', labeled_layout_tokens)
        semantic_content_iterable = (
            self.reference_segmenter_model
            .iter_semantic_content_for_labeled_layout_tokens(labeled_layout_tokens)
        )
        reference_list = SemanticReferenceList(list(semantic_content_iterable))
        semantic_document.back_section.add_content(reference_list)

    def _iter_parse_semantic_references(
        self,
        semantic_raw_references: List[SemanticRawReference]
    ) -> Iterable[Union[SemanticReference, SemanticInvalidReference]]:
        layout_documents = [
            LayoutDocument.for_blocks([semantic_raw_reference.merged_block])
            for semantic_raw_reference in semantic_raw_references
        ]
        labeled_layout_tokens_list = (
            self.citation_model
            .predict_labels_for_layout_documents(
                layout_documents,
                app_features_context=self.app_features_context
            )
        )
        LOGGER.debug('labeled_layout_tokens_list: %r', labeled_layout_tokens_list)
        for labeled_layout_tokens, semantic_raw_reference in zip(
            labeled_layout_tokens_list, semantic_raw_references
        ):
            semantic_content_iterable = (
                self.citation_model
                .iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens,
                    semantic_raw_reference=semantic_raw_reference
                )
            )
            ref: Optional[Union[SemanticReference, SemanticInvalidReference]] = None
            for semantic_content in semantic_content_iterable:
                if isinstance(semantic_content, (SemanticReference, SemanticInvalidReference)):
                    ref = semantic_content
            if not ref:
                raise AssertionError('no semantic reference extracted')
            yield ref

    def _extract_reference_fields_from_raw_references(
        self,
        semantic_document: SemanticDocument
    ):
        reference_lists = list(semantic_document.back_section.iter_by_type(
            SemanticReferenceList
        ))
        semantic_raw_references = [
            raw_reference
            for reference_list in reference_lists
            for raw_reference in reference_list.iter_by_type(SemanticRawReference)
        ]
        semantic_references = list(self._iter_parse_semantic_references(
            semantic_raw_references
        ))
        LOGGER.debug('semantic_references: %r', semantic_references)
        semantic_reference_by_semantic_raw_reference_id = {
            id(semantic_raw_reference): semantic_reference
            for semantic_raw_reference, semantic_reference in zip(
                semantic_raw_references, semantic_references
            )
        }
        LOGGER.debug(
            'semantic_reference_by_semantic_raw_reference_id keys: %s',
            semantic_reference_by_semantic_raw_reference_id.keys()
        )
        for reference_list in reference_lists:
            updated_content: List[SemanticContentWrapper] = []
            for semantic_content in reference_list:
                if isinstance(semantic_content, SemanticRawReference):
                    semantic_reference = semantic_reference_by_semantic_raw_reference_id[
                        id(semantic_content)
                    ]
                    updated_content.append(semantic_reference)
                    continue
                updated_content.append(semantic_content)
            reference_list.mixed_content = updated_content

    def _iter_parse_semantic_name_lists(
        self,
        semantic_raw_name_lists: Sequence[T_SemanticRawNameList],
        name_type: Type[T_SemanticName]
    ) -> Iterable[Tuple[T_SemanticRawNameList, List[SemanticContentWrapper]]]:
        layout_documents = [
            LayoutDocument.for_blocks([semantic_raw_name_list.merged_block])
            for semantic_raw_name_list in semantic_raw_name_lists
        ]
        labeled_layout_tokens_list = (
            self.name_citation_model
            .predict_labels_for_layout_documents(
                layout_documents,
                app_features_context=self.app_features_context
            )
        )
        LOGGER.debug('labeled_layout_tokens_list: %r', labeled_layout_tokens_list)
        for labeled_layout_tokens, semantic_raw_name_list in zip(
            labeled_layout_tokens_list, semantic_raw_name_lists
        ):
            semantic_content_iterable = (
                self.name_citation_model
                .iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens,
                    name_type=name_type
                )
            )
            yield semantic_raw_name_list, list(semantic_content_iterable)

    def _extract_reference_name_lists_from_raw_references(
        self,
        semantic_document: SemanticDocument
    ):
        reference_lists = list(semantic_document.back_section.iter_by_type(
            SemanticReferenceList
        ))
        ref_list = [
            ref
            for reference_list in reference_lists
            for ref in reference_list.iter_by_type(SemanticReference)
        ]
        if self.config.extract_citation_authors:
            raw_authors = [
                raw_author
                for ref in ref_list
                for raw_author in ref.iter_by_type(SemanticRawAuthors)
            ]
        else:
            raw_authors = []
        if self.config.extract_citation_editors:
            raw_editors = [
                raw_author
                for ref in ref_list
                for raw_author in ref.iter_by_type(SemanticRawEditors)
            ]
        else:
            raw_editors = []
        content_list_by_raw_author_id = {
            id(raw_author): content_list
            for raw_author, content_list in (
                self._iter_parse_semantic_name_lists(raw_authors, name_type=SemanticAuthor)
            )
        }
        content_list_by_raw_editor_id = {
            id(raw_author): content_list
            for raw_author, content_list in (
                self._iter_parse_semantic_name_lists(raw_editors, name_type=SemanticEditor)
            )
        }
        LOGGER.debug(
            'content_list_by_raw_author_id keys: %s',
            content_list_by_raw_author_id.keys()
        )
        LOGGER.debug(
            'content_list_by_raw_editor_id keys: %s',
            content_list_by_raw_editor_id.keys()
        )
        for reference_list in reference_lists:
            for semantic_content in reference_list:
                if isinstance(semantic_content, SemanticReference):
                    if self.config.extract_citation_authors:
                        semantic_content.flat_map_inplace_by_type(
                            SemanticRawAuthors,
                            lambda raw_author: content_list_by_raw_author_id[
                                id(raw_author)
                            ]
                        )
                    if self.config.extract_citation_editors:
                        semantic_content.flat_map_inplace_by_type(
                            SemanticRawEditors,
                            lambda raw_editor: content_list_by_raw_editor_id[
                                id(raw_editor)
                            ]
                        )

    def _iter_parse_semantic_content_lists(
        self,
        semantic_raw_content_lists: Sequence[T_SemanticContentWrapper],
        model: Model
    ) -> Iterable[Tuple[T_SemanticContentWrapper, List[SemanticContentWrapper]]]:
        layout_documents = [
            LayoutDocument.for_blocks([semantic_raw_name_list.merged_block])
            for semantic_raw_name_list in semantic_raw_content_lists
        ]
        labeled_layout_tokens_list = (
            model
            .predict_labels_for_layout_documents(
                layout_documents,
                app_features_context=self.app_features_context
            )
        )
        LOGGER.debug('labeled_layout_tokens_list: %r', labeled_layout_tokens_list)
        for labeled_layout_tokens, semantic_raw_name_list in zip(
            labeled_layout_tokens_list, semantic_raw_content_lists
        ):
            semantic_content_iterable = (
                model
                .iter_semantic_content_for_labeled_layout_tokens(
                    labeled_layout_tokens
                )
            )
            yield semantic_raw_name_list, list(semantic_content_iterable)

    def _extract_semantic_content_from_raw_content(
        self,
        semantic_document: SemanticDocument,
        semantic_type: Type[T_SemanticContentWrapper],
        model: Model
    ):
        parents = [
            parent
            for root in [
                semantic_document.body_section,
                semantic_document.back_section
            ]
            for parent in root.iter_parent_by_semantic_type_recursively(
                semantic_type
            )
        ]
        raw_content_lists = [
            raw_content
            for parent in parents
            for raw_content in parent.iter_by_type(semantic_type)
        ]
        content_list_by_raw_content_id = {
            id(raw_content): content_list
            for raw_content, content_list in (
                self._iter_parse_semantic_content_lists(
                    raw_content_lists,
                    model
                )
            )
        }
        LOGGER.debug(
            'content_list_by_raw_content_id keys: %s',
            content_list_by_raw_content_id.keys()
        )
        for parent in parents:
            parent.flat_map_inplace_by_type(
                semantic_type,
                lambda raw_content: content_list_by_raw_content_id[
                    id(raw_content)
                ]
            )

    def _extract_figure_fields_from_raw_figures(
        self,
        semantic_document: SemanticDocument
    ):
        self._extract_semantic_content_from_raw_content(
            semantic_document,
            SemanticRawFigure,
            self.figure_model
        )

    def _extract_table_fields_from_raw_tables(
        self,
        semantic_document: SemanticDocument
    ):
        self._extract_semantic_content_from_raw_content(
            semantic_document,
            SemanticRawTable,
            self.table_model
        )

    def _update_semantic_section_using_segmentation_result_and_fulltext_model(
        self,
        semantic_section: SemanticSection,
        segmentation_label_result: LayoutDocumentLabelResult,
        segmentation_tag: str,
        section_type: str
    ):
        layout_document = segmentation_label_result.get_filtered_document_by_label(
            segmentation_tag
        ).remove_empty_blocks()
        self._update_semantic_section_using_layout_document_and_fulltext_model(
            semantic_section,
            layout_document,
            section_name=segmentation_tag,
            section_type=section_type
        )

    def _update_semantic_section_using_layout_document_and_fulltext_model(
        self,
        semantic_section: SemanticSection,
        layout_document: LayoutDocument,
        section_name: str,
        section_type: str
    ):
        LOGGER.debug('layout_document (%r): %s', section_name, layout_document)
        if not layout_document.pages:
            return
        labeled_layout_tokens = self.fulltext_model.predict_labels_for_layout_document(
            layout_document,
            app_features_context=self.app_features_context
        )
        LOGGER.debug('labeled_layout_tokens (%r): %r', section_name, labeled_layout_tokens)
        entity_blocks = self.fulltext_model.iter_entity_layout_blocks_for_labeled_layout_tokens(
            labeled_layout_tokens
        )
        self.fulltext_model.update_section_with_entity_blocks(
            semantic_section,
            entity_blocks,
            section_type=section_type
        )

    def get_tei_document_for_layout_document(
        self,
        layout_document: LayoutDocument
    ) -> TeiDocument:
        return get_tei_for_semantic_document(
            self.get_semantic_document_for_layout_document(
                layout_document
            )
        )



import logging
from typing import Callable, Optional, Sequence, TypeVar

from flask import request
from werkzeug.exceptions import BadRequest

from sciencebeam_parser.utils.media_types import (
    get_first_matching_media_type
)
from sciencebeam_parser.utils.data_wrapper import MediaDataWrapper


LOGGER = logging.getLogger(__name__)


T = TypeVar('T')


DEFAULT_FILENAME = 'file'


def get_optional_post_data_wrapper() -> MediaDataWrapper:
    if not request.files:
        return MediaDataWrapper(
            data=request.data,
            media_type=request.mimetype,
            filename=request.args.get('filename')
        )
    supported_file_keys = ['file', 'input']
    for name in supported_file_keys:
        if name not in request.files:
            continue
        uploaded_file = request.files[name]
        data = uploaded_file.stream.read()
        return MediaDataWrapper(
            data=data,
            media_type=uploaded_file.mimetype,
            filename=uploaded_file.filename
        )
    raise BadRequest(
        f'missing file named one pf "{supported_file_keys}", found: {request.files.keys()}'
    )


def get_required_post_data_wrapper() -> MediaDataWrapper:
    data_wrapper = get_optional_post_data_wrapper()
    if not data_wrapper.data:
        raise BadRequest('no contents')
    return data_wrapper


def get_required_post_data() -> bytes:
    return get_required_post_data_wrapper().data


def get_request_accept_media_types() -> Sequence[str]:
    accept_media_types = list(request.accept_mimetypes.values())
    LOGGER.info('accept_media_types: %s', accept_media_types)
    return accept_media_types


def assert_and_get_first_matching_media_type(
    accept_media_types: Sequence[str],
    available_media_types: Sequence[str]
) -> str:
    media_type = get_first_matching_media_type(
        accept_media_types,
        available_media_types
    )
    if not media_type:
        raise BadRequest(
            f'unsupported accept media types: {accept_media_types},'
            f' supported types are: {available_media_types}'
        )
    LOGGER.info('resolved media type: %r', media_type)
    return media_type


def assert_and_get_first_accept_matching_media_type(
    available_media_types: Sequence[str]
) -> str:
    return assert_and_get_first_matching_media_type(
        get_request_accept_media_types(),
        available_media_types
    )


def get_typed_request_arg(
    name: str,
    type_: Callable[[str], T],
    default_value: Optional[T] = None,
    required: bool = False
) -> Optional[T]:
    value = request.args.get(name)
    if value:
        return type_(value)
    if required:
        raise ValueError(f'request arg {name} is required')
    return default_value


def str_to_bool(value: str) -> bool:
    value_lower = value.lower()
    if value_lower in {'true', '1'}:
        return True
    if value_lower in {'false', '0'}:
        return False
    raise ValueError('unrecognised boolean value: %r' % value)


def get_bool_request_arg(
    name: str,
    default_value: Optional[bool] = None,
    required: bool = False
) -> Optional[bool]:
    return get_typed_request_arg(
        name, str_to_bool, default_value=default_value, required=required
    )


def get_int_request_arg(
    name: str,
    default_value: Optional[int] = None,
    required: bool = False
) -> Optional[int]:
    return get_typed_request_arg(
        name, int, default_value=default_value, required=required
    )


def get_str_request_arg(
    name: str,
    default_value: Optional[str] = None,
    required: bool = False
) -> Optional[str]:
    return get_typed_request_arg(
        name, str, default_value=default_value, required=required
    )



import os
from typing import NamedTuple, Sequence

import subprocess


class UnoOfficePaths(NamedTuple):
    uno_path: str
    uno_python_path: str


class EnvironmentVariables:
    UNO_PATH = 'UNO_PATH'
    UNO_PYTHON_PATH = 'UNO_PYTHON_PATH'
    UNO_OFFICE_BINARY_PATH = 'UNO_OFFICE_BINARY_PATH'


class DefaultValues:
    UNO_PATH = '/usr/lib/python3/dist-packages'
    UNO_PYTHON_PATH = 'python3'
    UNO_OFFICE_BINARY_PATH = '/usr/lib/libreoffice/program/soffice.bin'


def get_uno_path() -> str:
    return os.environ.get('UNO_PATH') or DefaultValues.UNO_PATH


def get_uno_python_path() -> str:
    return os.environ.get('UNO_PYTHON_PATH') or DefaultValues.UNO_PYTHON_PATH


def get_uno_office_binary_path() -> str:
    return os.environ.get('UNO_OFFICE_BINARY_PATH') or DefaultValues.UNO_OFFICE_BINARY_PATH


def find_offices() -> Sequence[UnoOfficePaths]:
    return [UnoOfficePaths(
        uno_path=get_uno_path(),
        uno_python_path=get_uno_python_path()
    )]


def find_pyuno_office() -> UnoOfficePaths:
    offices = find_offices()
    if not offices:
        raise RuntimeError('no suitable office installation found')
    for office in offices:
        try:
            subprocess.check_output(
                [office.uno_python_path, '-c', 'import uno, unohelper'],
                env={'PYTHONPATH': office.uno_path}
            )
            return office
        except subprocess.CalledProcessError:
            pass
        except OSError:
            pass
    raise RuntimeError(
        'none of the potential office installations seem to function, tried: %s' % offices
    )


def get_start_listener_command(port: int) -> Sequence[str]:
    return [
        get_uno_office_binary_path(),
        '--headless',
        '--invisible',
        '--nocrashreport',
        '--nodefault',
        '--nofirststartwizard',
        '--nologo',
        '--norestore',
        '--accept=socket,host=localhost,port={port};urp;StarOffice.ServiceManager'.format(
            port=port
        )
    ]
